# -*- coding: utf-8 -*-

"""Heterogeneity models for sequence evolution.

References
----------
.. [1] Z. Yang.
   Computational molecular evolution.
   Oxford series in ecology and evolution. Oxford University Press, 2006.
   ISBN 978-0-19-856699-1 978-0-19-856702-8.
"""

import numpy as np

from asymmetree.seqevolve.EvolvingSequence import State


__author__ = 'David Schaller'


class HetModel:
    """Heterogeneity model.
    
    Supports rate heterogeneity based on a Gamma distribution ('+Gamma'-model)
    as well as invariant sites ('+I'-model), see [1].
    
    References
    ----------
    .. [1] Z. Yang.
       Computational molecular evolution.
       Oxford series in ecology and evolution. Oxford University Press, 2006.
       ISBN 978-0-19-856699-1 978-0-19-856702-8.
    """
    
    def __init__(self, alpha, classes=5, sitewise=False, invariant=0.0):
        """
        Parameters
        ----------
        alpha : float
            Parameter of the Gamma distribution (with mean 1) from which the
            rate factors are drawn.
        classes : int, optional
            Number of classes such that sites in the same class share the same
            rate factor. The default is 5.
        sitewise : bool, optional
            If True, ignore the 'classes' attribute and treat each site as its
            own class. The default is False.
        invariant : float, optional
            Expected proportion of invariant sites, i.e., sites where no
            mutations happen. The default is 0.0.
        """
        
        if not isinstance(alpha, float) or alpha <= 0.0:
            raise ValueError('heterogeneity parameter alpha must be a float > 0.0')
        self._alpha = alpha
        
        if not isinstance(classes, int) or classes <= 0:
            raise ValueError('number of classes must be an int >0')
        self.classes = classes
        
        self.sitewise = sitewise
        
        if not isinstance(invariant, float) or invariant < 0.0 or invariant > 1.0:
            raise ValueError('proportion of invariant sites must be in [0.0, 1.0]')
        self._invariant = invariant
        
        if not self.sitewise:
            self._initilize_classes()
        
    
    def _initilize_classes(self):
        
        if self.classes == 1:
            self._class_rates = [1.0]
            
        else:
            self._class_rates = np.random.gamma(self._alpha,
                                                scale=1/self._alpha,
                                                size=self.classes)
        
        
    def assign(self, sequence, exclude_inherited=True):
        """Assign rate classes and rate factors tothe sites of a sequence.
        
        Parameters
        ----------
        sequence : asymmetree.seqevolve.EvolvingSequence.EvoSeq
            The sequence.
        exclude_inherited : bool, optional
            If True, do not assign new classes and rates to inherited sites,
            i.e., only if the sequence corresponds to the root of a tree along
            which is simulated or if new sites are added as a result of an
            insertion event.
        """
        
        n = len(sequence)
        if exclude_inherited:
            n -= sequence.count_status(State.INHERITED)
        
        rate_classes, rate_factors = self._draw(n)
        
        pos = 0
        for site in sequence:
            
            if not exclude_inherited or site.status != State.INHERITED:
                site.rate_class = rate_classes[pos] if rate_classes[pos] is not None else site.site_id
                site.rate_factor = rate_factors[pos]
                pos += 1
                    
                    
    def _draw(self, n):
        
        # mode 1: sitewise heterogeneity
        if self.sitewise:
            
            rate_classes = [None for _ in range(n)]
            rate_factors = np.random.gamma(self._alpha, scale=1/self._alpha, size=n)
        
        # mode 2: one or multiple  classes
        else:
            
            rate_classes = np.random.randint(self.classes, size=n).tolist()
            rate_factors = np.asarray([self._class_rates[c] for c in rate_classes])
            
        if self._invariant:
            
            drawn_variable = np.random.random(n) > self._invariant
            
            for i in range(n):
                if not drawn_variable[i]:
                    rate_classes[i] = 'invariant'
            
            rate_factors *= drawn_variable
        
        return rate_classes, rate_factors


# -*- coding: utf-8 -*-

"""Constructing the true alignment of simulated sequences."""

import numpy as np
import networkx as nx


__author__ = 'David Schaller'


class AlignmentBuilder:
    """Construction of the true alignment of simulated sequences."""
    
    def __init__(self, tree, sequence_dict, alphabet, include_inner=True):
        """
        Parameters
        ----------
        tree : Tree
            The tree along which the sequences where simulated.
        sequence_dict : dict
            A dict containing the TreeNode instances in the tree as keys and 
            the simulated sequences as values (instances of type EvoSeq).
        alphabet
            The alphabet of the substitution model that was used.
        include_inner : bool, optional
            If True, the alignment also contains the sequences of the inner
            nodes of the tree. The default is True.
        """
        
        self.tree = tree
        self.sequence_dict = sequence_dict
        self.include_inner = include_inner
        
        if alphabet[-1] == '-':
            self.alphabet = alphabet
        else:
            self.alphabet = alphabet + '-'
        
        
    def build(self):
        """Build the true alignment.
        
        Returns
        -------
        dict
            A dict containing the TreeNode instances in the tree as keys and 
            the str sequences as values that include the necessary gaps.
        """
        
        self._get_preorder()
        self._sort_sites()
        self._alignment_matrix()
        
        return self._sequences()
        
    
    def _get_preorder(self):
        
        if self.include_inner:
            self.nodes = [v for v in self.tree.preorder()]
        else:
            self.nodes = [v for v in self.tree.preorder() if not v.children]
            
        self.node_index = {item: index for index, item in enumerate(self.nodes)}
        
    
    def _sort_sites(self):
        
        G = nx.DiGraph()
        
        for node in self.nodes:
            
            seq = self.sequence_dict[node]
            
            # handle case |seq| = 1 correctly
            if len(seq) == 1:
                G.add_node(seq[0].site_id)
            
            # add nodes and edges to auxiliary graph
            for first, second in seq.element_pairs():
                G.add_edge(first.site_id, second.site_id)
                
        self.sites = list(nx.topological_sort(G))
                
    
    def _alignment_matrix(self):
        
        self.alignment = np.zeros((len(self.nodes), len(self.sites)),
                             dtype=np.int8)
        positions = [self.sequence_dict[v]._first for v in self.nodes]
        
        for j in range(len(self.sites)):
            for i in range(len(positions)):
                
                if positions[i] and self.sites[j] == positions[i].site_id:
                    self.alignment[i, j] = positions[i]._value
                    positions[i] = positions[i]._next
                else:
                    self.alignment[i, j] = -1
    
    
    def _sequences(self):
        
        result = {}
        
        for i in range(len(self.nodes)):
            sequence = "".join(self.alphabet[x] for x in self.alignment[i, :])
            result[self.nodes[i]] = sequence
        
        return result


# -*- coding: utf-8 -*-

"""
Simulation of dated species trees.
"""

import random, sys
import numpy as np

from tralda.datastructures.Tree import Tree, TreeNode

from asymmetree.tools.PhyloTreeTools import (delete_losses_and_contract,
                                             remove_planted_root,
                                             distance_from_timing)


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                         USER INTERFACE FUNCTION
# --------------------------------------------------------------------------

def species_tree_n(n, 
                   model='yule',
                   innovation=False,
                   planted=True,
                   remove_extinct=False,
                   **kwargs):
    """Simulates a species tree S with n leaves.
    
    Parameters
    ----------
    n : int
        Number of leaves in the resulting tree that correspond to extant
        species.
    model : str, optional
        Simulation model to be applied, the default is 'yule', see [1].
        Other available model are birth-death process ('BDP', [2]), and
        episodic birth-death process ('EBDP', [3]).
    innovation : bool, optional
        If True, use the innovation model, see [4], to sample a lineage for
        the next speciation event. Only available for the Yule model.
        The default is False, in which case the lineage is chosen uniformly at
        random among the currently existing lineages.
    planted : bool, optional
        Add a planted root that has the canonical root as its single neighbor; 
        default is True.
    remove_extinct : bool, optional
        Remove all branches that lead to extinctions, only relevant for some
        models; default is False.
    birth_rate : float, optional
        The birth rate for models such as 'yule' and 'BDP'.
    death_rate : float, optional
        The death rate for models such as 'BDP'.
    episodes : list, optional
        The episodes for the model 'EBDP'.
    contraction_probability : float, optional
        Probability that an inner edge is contracted. The default is 0.0, in
        which case the tree is binary. Only one of this parameter and
        'contraction_proportion' may be non-zero.
    contraction_proportion : float, optional
        The proportion of inner edges to be contracted. The default is 0.0, in
        which case the tree is binary. Only one of this parameter and 
        'contraction_probabilty' may be non-zero.
    contraction_bias : str, optional
        Specifies whether shorter edges, i.e., with a smaller difference t of
        the time stamps, have a higher probability to be contracted. Only
        relevant if 'contraction_proportion' > 0.0. The default is False,
        in which case all edges have the same probability to be contracted.
        The options 'inverse' and 'exponential' mean that an edge is sampled
        weighted by 1/t or e^(-t), respectively.
    bias_strength : float, optional
        Intensity factor for preferring shorter edges to be contracted. The
        default is 1.0.
        
    Raises
    ------
    ValueError
        If unknown parameter values are passed.
        
    Returns
    -------
    Tree
        The simulated species tree.
        
    References
    ----------
    .. [1] G. U. Yule.
       A mathematical theory of evolution, based on the conclusions of Dr. J. 
       C.Willis, F. R. S. 
       In: Phil. Trans. R. Soc. Lond. B, 1924, 213, 21–87.
       doi:10.1098/rstb.1925.0002.
    .. [2] D. G. Kendall.
       On the Generalized "Birth-and-Death" Process.
       In: Ann. Math. Statist. 1948, 19, 1–15.
       doi:10.1214/aoms/1177730285.
    .. [3] T. Stadler.
       Simulating trees with a fixed number of extant species.
       In: Syst. Biol. 2011, 60, 676–684.
       doi:10.1093/sysbio/syr029.
    .. [4] S. Keller-Schmidt, K. Klemm.
       A model of macroevolution as a branching process based on innovations.
       In: Adv. Complex Syst.,2012, 15, 1250043.
       doi:10.1142/S0219525912500439.
    """
    
    # parameter checking
    if not isinstance(n, int) or n < 0:
        raise ValueError('n must be an int >=0')
    elif n == 0:
        return Tree(None)
    
    if not isinstance(model, str):
        raise ValueError("model must be of type 'str'")
    
    # choice of the main simulation algorithm
    if model.lower() == 'yule':
        tree = _yule_n(n, kwargs.get('birth_rate'), innovation)
    elif model.upper() == 'BDP':
        tree = _BDP_n(n, **kwargs)
    elif model.upper() == 'EBDP':
        tree = _EBDP_n(n, **kwargs)
    else:
        raise ValueError(f"model '{model}' is not available")
        
    # remove extinct branches for models that include losses
    if remove_extinct and model.upper() in ('BDP', 'EBDP'):
        delete_losses_and_contract(tree, inplace=True)
        
    # remove planted edge for models that are planted by construction
    if not planted and model.upper() in ('YULE', 'BDP', 'EBDP'):
        remove_planted_root(tree, inplace=True)
    
    # make tree non_binary by random contraction of edges
    nonbinary(tree, inplace=True, **kwargs)
    
    # assign the distance attribute to all vertices
    distance_from_timing(tree)
        
    return tree


def species_tree_age(age, 
                     model='yule', 
                     innovation=False,
                     **kwargs):
    """Simulates a (planted) species tree S of the specified age.
    
    Parameters
    ----------
    age : float
        Simulation time, i.e., the time span from the root of the tree to the
        leaves that correspond to extant species.
    model : str, optional
        Simulation model to be applied, the default is 'yule', see [1].
        Other available model are birth-death process ('BDP', [2]), and
        episodic birth-death process ('EBDP', [3]).
    innovation : bool, optional
        If True, use the innovation model, see [4], to sample a lineage for
        the next speciation event. The default is False, in which case the
        lineage is chosen uniformly at random among the currently existing
        lineages.
    birth_rate : float, optional
        The birth rate for models such as 'yule' and 'BDP'.
    death_rate : float, optional
        The death rate for models such as 'BDP'.
    episodes : list, optional
        The episodes for the model 'EBDP'.
    contraction_probability : float, optional
        Probability that an inner edge is contracted. The default is 0.0, in
        which case the tree is binary. Only one of this parameter and
        'contraction_proportion' may be non-zero.
    contraction_proportion : float, optional
        The proportion of inner edges to be contracted. The default is 0.0, in
        which case the tree is binary. Only one of this parameter and 
        'contraction_probabilty' may be non-zero.
    contraction_bias : str, optional
        Specifies whether shorter edges, i.e., with a smaller difference t of
        the time stamps, have a higher probability to be contracted. Only
        relevant if 'contraction_proportion' > 0.0. The default is False,
        in which case all edges have the same probability to be contracted.
        The options 'inverse' and 'exponential' mean that an edge is sampled
        weighted by 1/(a * t) or e^(-a * t), respectively, where a is a
        user-defined factor.
    bias_strength : float, optional
        Intensity factor for preferring shorter edges to be contracted. The
        default is 1.0.
        
    Raises
    ------
    ValueError
        If unknown or invalid parameter values are passed.
        
    Returns
    -------
    Tree
        The simulated species tree.
        
    References
    ----------
    .. [1] G. U. Yule.
       A mathematical theory of evolution, based on the conclusions of Dr. J. 
       C.Willis, F. R. S. 
       In: Phil. Trans. R. Soc. Lond. B, 1924, 213, 21–87.
       doi:10.1098/rstb.1925.0002.
    .. [2] D. G. Kendall.
       On the Generalized "Birth-and-Death" Process.
       In: Ann. Math. Statist. 1948, 19, 1–15.
       doi:10.1214/aoms/1177730285.
    .. [3] T. Stadler.
       Simulating trees with a fixed number of extant species.
       In: Syst. Biol. 2011, 60, 676–684.
       doi:10.1093/sysbio/syr029.
    .. [4] S. Keller-Schmidt, K. Klemm.
       A model of macroevolution as a branching process based on innovations.
       In: Adv. Complex Syst.,2012, 15, 1250043.
       doi:10.1142/S0219525912500439.
    """
    
    # parameter checking
    if not isinstance(age, (float, int)) or age <= 0.0:
        raise ValueError('age must be a number >0')
    elif isinstance(age, int):
        age = float(age)
        
    if not isinstance(model, str):
        raise ValueError("model must be of type 'str'")
    
    # main simulation algorithm
    if model.lower() == 'yule':
        tree = _yule_age(age, kwargs.get('birth_rate'), innovation)
    elif model.upper() == 'BDP':
        tree = _BDP_age(age, innovation, **kwargs)
    elif model.upper() == 'EBDP':
        tree = _EBDP_age(age, innovation, **kwargs)
    else:
        raise ValueError(f"model '{model}' is not available")
        
    # make tree non_binary by random contraction of edges
    nonbinary(tree, inplace=True, **kwargs)
    
    # assign the distance attribute to all vertices
    distance_from_timing(tree)
        
    return tree


def species_tree_n_age(n, age, 
                       model='yule',
                       innovation=False,
                       birth_rate=1.0,
                       death_rate=0.0,
                       **kwargs):
    """Simulate a (planted) species tree S with n leaves and of the specified
    age.
    
    The tree is sampled under the Yule or BDP model conditioned on the number
    n of surviving species and with a given age [1].
    The resulting tree does not contain loss leaves even if the specified death
    rate is > 0 as a consequence of the tree sampling method.
    
    Parameters
    ----------
    n : int
        Number of leaves in the resulting tree that correspond to extant
        species.
    age : float
        Simulation time, i.e., the time span from the root of the tree to the
        leaves that correspond to extant species.
    model : str, optional
        Simulation model to be applied, the default is 'yule', see [2].
        Alternatively the birth-death process ('BDP', [3]) is available.
    innovation : bool, optional
        If True, use the innovation model, see [4], to sample a lineage for
        the next speciation event. The default is False, in which case the
        lineage is chosen uniformly at random among the currently existing
        lineages.
    birth_rate : float, optional
        The birth rate for models 'yule' and 'BDP'. The default is 1.0.
    death_rate : float, optional
        The death rate for model 'BDP'. The deafult is 0.0.
    contraction_probability : float, optional
        Probability that an inner edge is contracted. The default is 0.0, in
        which case the tree is binary. Only one of this parameter and
        'contraction_proportion' may be non-zero.
    contraction_proportion : float, optional
        The proportion of inner edges to be contracted. The default is 0.0, in
        which case the tree is binary. Only one of this parameter and 
        'contraction_probabilty' may be non-zero.
    contraction_bias : str, optional
        Specifies whether shorter edges, i.e., with a smaller difference t of
        the time stamps, have a higher probability to be contracted. Only
        relevant if 'contraction_proportion' > 0.0. The default is False,
        in which case all edges have the same probability to be contracted.
        The options 'inverse' and 'exponential' mean that an edge is sampled
        weighted by 1/(a * t) or e^(-a * t), respectively, where a is a
        user-defined factor.
    bias_strength : float, optional
        Intensity factor for preferring shorter edges to be contracted. The
        default is 1.0.
        
    Raises
    ------
    ValueError
        If unknown or invalid parameter values are passed.
        
    Returns
    -------
    Tree
        The simulated species tree.
        
    References
    ----------
    .. [1] T. Stadler.
       Simulating trees with a fixed number of extant species.
       In: Syst. Biol. 2011, 60, 676–684.
       doi:10.1093/sysbio/syr029.
    .. [2] G. U. Yule.
       A mathematical theory of evolution, based on the conclusions of Dr. J. 
       C.Willis, F. R. S. 
       In: Phil. Trans. R. Soc. Lond. B, 1924, 213, 21–87.
       doi:10.1098/rstb.1925.0002.
    .. [3] D. G. Kendall.
       On the Generalized "Birth-and-Death" Process.
       In: Ann. Math. Statist. 1948, 19, 1–15.
       doi:10.1214/aoms/1177730285.
    .. [4] S. Keller-Schmidt, K. Klemm.
       A model of macroevolution as a branching process based on innovations.
       In: Adv. Complex Syst.,2012, 15, 1250043.
       doi:10.1142/S0219525912500439.
    """
    
    # parameter checking
    if not isinstance(age, (float, int)) or age <= 0.0:
        raise ValueError('age must be a number >0')
    elif isinstance(age, int):
        age = float(age)
        
    if not isinstance(n, int) or n < 0:
        raise ValueError('n must be an int >=0')
    elif n == 0:
        return Tree(None)
        
    if not isinstance(model, str):
        raise ValueError("model must be of type 'str'")
    
    # main simulation algorithm
    if model.lower() == 'yule':
        tree = _yule_n_age(n, age, kwargs.get('birth_rate'), innovation)
    elif model.upper() == 'BDP':
        tree = _BDP_n_age(n, age,
                          kwargs.get('birth_rate'), kwargs.get('death_rate'),
                          innovation)
    else:
        raise ValueError(f"model '{model}' is not available")
        
    # make tree non_binary by random contraction of edges
    nonbinary(tree, inplace=True, **kwargs)
    
    # assign the distance attribute to all vertices
    distance_from_timing(tree)
        
    return tree


def nonbinary(tree,
              contraction_probability=0.0,
              contraction_proportion=0.0,
              contraction_bias=False,
              bias_strength=1.0,
              inplace=False,
              **kwargs):
    """Introduce multifurcation into a tree by contraction of inner edges.
    
    Parameters
    ----------
    tree : Tree
        The tree whose edges shall be contracted.
    contraction_probability : float, optional
        Probability that an inner edge is contracted; results in non-binary 
        tree; default is 0.0. Only one of this parameter and 
        'contraction_proportion' may be non-zero.
    contraction_proportion : float, optional
        The proportion of inner edges to be contracted; results in non-binary 
        tree; default is 0.0. Only one of this parameter and 
        'contraction_probabilty' may be non-zero.
    contraction_bias : str, optional
        Specifies whether shorter edges, i.e., with a smaller difference t of
        the time stamps, have a higher probability to be contracted. Only
        relevant if 'contraction_proportion' > 0.0. The default is False,
        in which case all edges have the same probability to be contracted.
        The options 'inverse' and 'exponential' mean that an edge is sampled
        weighted by 1/(a * t) or e^(-a * t), respectively, where a is a
        user-defined factor.
    bias_strength : float, optional
        Intensity factor for preferring shorter edges to be contracted. The
        default is 1.0.
    inplace : bool, optional
        If True, the edges are contracted in the original tree instance;
        otherwise a copy of the tree in which edges are contracted is returned.
    
    Returns
    -------
    Tree
        The tree whose edges have been contracted according to the parameters.
    
    Raises
    ------
    ValueError
        If both contraction_probability and contraction_proportion are
        non-zero or do not lie in the interval [0, 1].
    """
    
    if contraction_probability != 0.0 and contraction_proportion != 0.0:
        raise ValueError("only one of parameters 'contraction_probability' "\
                         "and 'contraction_proportion' may be non-zero")
    
    if contraction_probability < 0.0 or contraction_probability > 1.0:
        raise ValueError('contraction probability must be in [0.0, 1.0]')
    
    if contraction_proportion < 0.0 or contraction_proportion > 1.0:
        raise ValueError('contraction proportion must be in [0.0, 1.0]')
    
    if not isinstance(bias_strength, (int, float)) or \
        bias_strength <= 0.0:
        raise ValueError('factor for contraction bias must be > 0')
    
    if not inplace:
        tree = tree.copy()
    
    edges = False
    if contraction_probability > 0.0:
         edges = _select_edges_by_probability(tree, contraction_probability,
                                              exclude_planted_edge=True)
         
    elif contraction_proportion > 0.0:
        edges = _select_edges_by_proportion(tree, contraction_proportion,
                                            contraction_bias, bias_strength,
                                            exclude_planted_edge=True)
    if edges:
        tree.contract(edges)
        distance_from_timing(tree)
    
    return tree


# --------------------------------------------------------------------------
#                         AUXILIARY FUNCTIONS 
# --------------------------------------------------------------------------


def _rescale(tree, height, inplace=True):
    
    if not inplace:
        tree = tree.copy()
    
    old_height = tree.root.tstamp
    
    # not available for trees that only consist of a root
    if old_height <= 0.0:
        raise RuntimeError("cannot rescale tree of "\
                           "height '{}'".format(old_height))
        
    scaling_factor = height / old_height
    
    for v in tree.preorder():
        v.tstamp *= scaling_factor
    
    return tree
            
            
def _select_edges_by_probability(tree, p, exclude_planted_edge=True):
    
    edges = []
    
    for u, v in tree.inner_edges():
        
        if exclude_planted_edge and (u is tree.root) and len(u.children) == 1:
            continue
        
        if random.random() < p:
            edges.append((u,v))
    
    return edges


def _select_edges_by_proportion(tree, p, weighting, weighting_factor,
                                exclude_planted_edge=True):
    
    edges = [e for e in tree.inner_edges()
             if not (exclude_planted_edge and (e[0] is tree.root) and 
                     len(e[0].children) == 1)]
    
    # number of edges to be sampled
    k = round(p * len(edges))
    
    if not weighting:
        return random.choices(edges, k=k)
    else:
        distances = [abs(u.tstamp-v.tstamp) for u, v in edges]
        if weighting == 'inverse':
            weights = 1 / (weighting_factor * np.asarray(distances))
        elif weighting == 'exponential':
            weights = np.exp(-weighting_factor * np.asarray(distances))
        else:
            raise ValueError(f'unknown mode for weighted sampling: {weighting}')
        
        return random.choices(edges, weights=weights, k=k)



def assign_losses(tree, proportion):
    """Randomly assigns a specified proportion of leaves as losses."""
    
    for leaf in tree.random_leaves(proportion):
        leaf.event = 'L'


# --------------------------------------------------------------------------
#                             FORWARD MODELS
# --------------------------------------------------------------------------

class _ForwardLineageSampler:
    """Sample lineages for speciation events (forward simulation).
    
    Optionally uses the innovation model [1] for sampling the lineage in which
    the next speciation occurs.
    
    References
    ----------
    .. [1] S. Keller-Schmidt, K. Klemm.
       A model of macroevolution as a branching process based on innovations.
       In: Adv. Complex Syst.,2012, 15, 1250043.
       doi:10.1142/S0219525912500439.
    """
    
    def __init__(self, innovation):
        
        self.innovation = innovation
        
        root = TreeNode(label=0, event=None, tstamp=0.0)
        self.tree = Tree(root)
        
        # lineages (branch id, parent node, feature set index)
        self.lineages = [(1, root, 0)]
        self.node_counter = 2
        
        if self.innovation:
            self.feature_counter = 0
            self.feature_sets = [frozenset([])]
            
            # feature set --> lineage index
            self.set_to_species = {frozenset([]): 0}
    
    
    def speciation(self, t):
        
        if self.innovation:
            
            loss_candidates = set()     # species for which loss of a feature 
                                        # can trigger a speciation
            for s in self.feature_sets:
                for f in s:
                    if s - {f} not in self.set_to_species:
                        loss_candidates.add(s)
            
            if loss_candidates: # speciation by oss of feature
            
                loss_candidates = list(loss_candidates)
                while True:
                    s = random.choice(loss_candidates)
                    f = np.random.randint(self.feature_counter)
                    new_s = s - {f}
                    if new_s not in self.set_to_species:
                        break
                    
            else:               # speciation by gain of feature (innovation)
                
                s = random.choice(self.feature_sets)
                new_feature = self.feature_counter
                self.feature_counter += 1
                new_s = s.union([new_feature])
            
            new_set_index = len(self.feature_sets)
            self.feature_sets.append(new_s)
            i = self.set_to_species[s]
            self.set_to_species[new_s] = len(self.lineages)
        
        else:
            i = np.random.randint(len(self.lineages))
            new_set_index = 0
        
        lin_id, parent, set_index = self.lineages[i]
        spec_node = TreeNode(label=lin_id, event='S', tstamp=t)
        parent.add_child(spec_node)
        
        self.lineages[i] = (self.node_counter, spec_node, set_index)
        self.lineages.append((self.node_counter+1, spec_node, new_set_index))
        self.node_counter += 2
        
        return spec_node
    
    
    def extinction(self, t):
        
        return self.lineage_extinction(np.random.randint(len(self.lineages)), t)
        
    
    def mass_extinction(self, surviving_rate, t):
        
        no_of_losses = round((1-surviving_rate) * len(self.lineages))
        
        # indices in decreasing order so that indices in self.lineages remain
        # valid upon removal
        chosen_losses = sorted(np.random.choice(len(self.lineages), 
                                                replace=False,
                                                size=no_of_losses),
                               reverse=True)
        
        for i in chosen_losses:
            self.lineage_extinction(i, t)
            
    
    def lineage_extinction(self, i, t):
        
        lin_id, parent, j = self.lineages[i]
        loss_node = TreeNode(label=lin_id, event='L', tstamp=t)
        parent.add_child(loss_node)
        self.lineages.pop(i)
        
        if self.innovation:
            feature_set = self.feature_sets[j]
            self.feature_sets.pop(j)
            del self.set_to_species[feature_set]
        
        return loss_node
    
    
    def finalize(self, time):
        
        for lin_id, parent, _ in self.lineages:
            parent.add_child( TreeNode(label=lin_id, event='S', tstamp=time) )
        
        for v in self.tree.preorder():
            v.tstamp = abs(v.tstamp - time)
        
        return self.tree


def _yule_n(n, birth_rate, innovation):
    
    if birth_rate is None:
        birth_rate = 1.0
    elif birth_rate <= 0.0:
        raise ValueError('birth rate must be >0')
    
    fls = _ForwardLineageSampler(innovation)
    t = 0.0
    
    while len(fls.lineages) < n:
        rate = len(fls.lineages) * birth_rate
        t += np.random.exponential(1/rate)
        fls.speciation(t)
        
    # add length for pendant lineages (cf. Hartmann et al. 2010)
    t += np.random.exponential(1/rate)
    
    return fls.finalize(t)


def _yule_age(age, birth_rate, innovation):
    
    if birth_rate is None:
        birth_rate = 1.0
    elif birth_rate <= 0.0:
        raise ValueError("birth rate must be >0")
    
    fls = _ForwardLineageSampler(innovation)
    t = 0.0
    
    while t < age:
        rate = len(fls.lineages) * birth_rate
        t += np.random.exponential(1/rate)
        if t >= age: break
        fls.speciation(t)
    
    return fls.finalize(age)


def _yule_n_age(n, age, birth_rate, innovation):
    
    return _BDP_n_age(n, age, birth_rate, 0.0, innovation)
    

def _BDP_age(age, innovation, **kwargs):
    
    # remove potentially supplied 'episodes' argument
    episodes = _EBDP_age_check_episodes(birth_rate = kwargs.get('birth_rate'),
                                        death_rate = kwargs.get('death_rate'))
    
    return _EBDP_age_forward(age, episodes, innovation)


def _BDP_n_age(n, age, birth_rate, death_rate, innovation):
    
    if birth_rate is None:
        birth_rate = 1.0
    elif birth_rate <= 0.0:
        raise ValueError('birth rate must be >0')
    if death_rate is None:
        death_rate = 0.0
    elif death_rate < 0.0:
        raise ValueError('death rate must be >=0')
    
    fls = _ForwardLineageSampler(innovation)
    inner_vertices = []
    
    while len(fls.lineages) < n:
        inner_vertices.append(fls.speciation(0.0))
    fls.finalize(0.0)
    
    tree = fls.tree
    tree.root.tstamp = age
    
    # the following code is adapated from TreeSim (Stadler 2011)
    
    spec_times = []
    rho = 1.0           # proportion of sampled extant species (for future
                        # extension of the function)
    random_uniform = np.random.random(len(inner_vertices))
    for r in random_uniform:
        
        lamb1 = rho * birth_rate
        mu1   = death_rate - birth_rate * (1 - rho)
        
        if birth_rate > death_rate:
            t = 1 / (lamb1 - mu1) * \
                np.log((lamb1-mu1 * np.exp((-lamb1+mu1) * age) - \
                        mu1*(1-np.exp((-lamb1+mu1)*age)) * r ) /
                       (lamb1 - mu1 * np.exp((-lamb1+mu1) * age) - 
                        lamb1 * (1-np.exp((-lamb1+mu1) * age)) * r)
                      )
        else:
            t = - ((age * r) / 
                   (-1 - birth_rate * rho * age + birth_rate * rho* age * r ))
            
        spec_times.append(t)
        
    spec_times.sort(reverse=True)
    
    for v, t in zip(inner_vertices, spec_times):
        v.tstamp = t
    
    return tree


def _EBDP_age(age, innovation, **kwargs):
    
    episodes = _EBDP_age_check_episodes(**kwargs)
    
    return _EBDP_age_forward(age, episodes, innovation)


def _EBDP_age_check_episodes(**kwargs):
    
    birth_rate = kwargs.get('birth_rate')
    death_rate = kwargs.get('death_rate')
    episodes = kwargs.get('episodes')
    
    # episodes parameter is prefered
    if episodes is not None:
        
        if len(episodes) == 0:
            raise ValueError("list of episodes must not be empty")
        
        for i in range(len(episodes)):
            
            if len(episodes[i]) != 4:
                raise ValueError("all episodes must contain 4 values: birth rate, "\
                                 "death rate, proportion of survivors, time stamp "\
                                 "(from recent time as 0)")
            
            birth_rate, death_rate, rho, t = episodes[i]
            if i == 0 and t != 0.0:
                raise ValueError("first episode must be at t=0.0")
            elif i > 0 and episodes[i-1][3] >= t:
                print(episodes[i-1][3], t)
                raise ValueError("episodes must be in correct temporal order")
            
            if birth_rate < 0.0 or death_rate < 0.0:
                raise ValueError("birth and death rate must be >=0 "\
                                 "in all episodes")
                
            if rho <= 0.0 or rho > 1.0:
                raise ValueError("proportion of survivors must be in (0.0, 1.0]")
        
        return episodes
    
    elif birth_rate is not None:
        
        if birth_rate < 0.0 or (death_rate is not None and death_rate < 0.0):
            raise ValueError("birth and death rate must be >=0")
        
        if death_rate is None:
            # default death rate = 0.0
            return [(birth_rate, 0.0, 1.0, 0.0)]
        else:
            return [(birth_rate, death_rate, 1.0, 0.0)]
        
    else:
        if death_rate:
            raise ValueError("birth rate (>=0) must be specified if death rate "\
                             "is supplied")
        
        # default birth rate = 1.0 and death rate = 0.0
        return [(1.0, 0.0, 1.0, 0.0)]
    

def _EBDP_age_forward(age, episodes, innovation):
    """Episodic birth–death process (EBDP), forward algorithm with max. age."""
    
    fls = _ForwardLineageSampler(innovation)
    t = 0.0  # current time (forward simulation, is reversed in the end)
    i = 0    # current episode
    
    # may lead to extinction of the single branch at time t=0
    fls.mass_extinction(episodes[i][2], episodes[i][3])
    
    while t < age:
        birth_rate, death_rate, *_ = episodes[i]
        
        rate = len(fls.lineages) * (birth_rate + death_rate)
        waiting_time = np.random.exponential(1/rate) if rate > 0.0 else float('inf')
        
        if i+1 < len(episodes) and t + waiting_time >= episodes[i+1][3]:
            fls.mass_extinction(episodes[i+1][2], episodes[i+1][3])
            t = episodes[i+1][3]
            i += 1
        
        elif t + waiting_time >= age:
            break
        
        else:
            t += waiting_time
            
            if birth_rate > np.random.uniform(low=0.0, 
                                              high=birth_rate+death_rate):
                fls.speciation(t)
            else:
                fls.extinction(t)
    
    return fls.finalize(age)


# --------------------------------------------------------------------------
#                            BACKWARD MODELS
# --------------------------------------------------------------------------


def _BDP_n(n, **kwargs):
    
    # remove potentially supplied 'episodes' argument
    episodes = _EBDP_check_episodes(birth_rate = kwargs.get('birth_rate'),
                                    death_rate = kwargs.get('death_rate'))
    
    return _EBDP_backward(n, episodes)


def _EBDP_n(n, **kwargs):
    
    episodes = _EBDP_check_episodes(**kwargs)
    
    return _EBDP_backward(n, episodes)


def _EBDP_check_episodes(**kwargs):
    
    birth_rate = kwargs.get('birth_rate')
    death_rate = kwargs.get('death_rate')
    episodes = kwargs.get('episodes')
    
    # episodes parameter is prefered
    if episodes is not None:
        
        if len(episodes) == 0:
            raise ValueError("list of episodes must not be empty")
        
        for i in range(len(episodes)):
            
            if len(episodes[i]) != 4:
                raise ValueError("all episodes must contain 4 values: birth rate, "\
                                 "death rate, proportion of survivors, time stamp "\
                                 "(from recent time as 0)")
            
            birth_rate, death_rate, rho, t = episodes[i]
            if i == 0 and t != 0.0:
                raise ValueError("first episode must be at t=0.0")
            elif i > 0 and episodes[i-1][3] >= t:
                print(episodes[i-1][3], t)
                raise ValueError("episodes must be in correct temporal order")
            
            if birth_rate <= 0.0 or birth_rate < death_rate:
                raise ValueError("birth rate must be >0 and >=death rate "\
                                 "in all episodes")
                
            if rho <= 0.0 or rho > 1.0:
                raise ValueError("proportion of survivors must be in (0.0, 1.0]")
        
        return episodes
    
    elif birth_rate is not None:
        
        if birth_rate <= 0.0 or (death_rate and birth_rate < death_rate):
            raise ValueError("birth rate must be >0 and >=death rate")
            
        if death_rate and death_rate < 0.0:
            raise ValueError("death rate must be >=0")
        
        if death_rate is None:
            return [(birth_rate, 0.0, 1.0, 0.0)]
        else:
            return [(birth_rate, death_rate, 1.0, 0.0)]
        
    else:
        if death_rate:
            raise ValueError("birth rate (>0) must be specified if death rate "\
                             "is supplied")
            
        return [(1.0, 0.0, 1.0, 0.0)]


def _EBDP_backward(n, episodes, max_tries=500):
    """Episodic birth–death process (EBDP).
    
    References
    ----------
    .. [1] T. Stadler.
       Simulating trees with a fixed number of extant species.
       In: Syst. Biol. 2011, 60, 676–684.
       doi:10.1093/sysbio/syr029.
    """
    
    birth_inv_sum = sum([1/episodes[i][0] for i in range(len(episodes))])
    
    for _ in range(max_tries):
        
        tree = None
        t = 0.0
        i = 0
        
        branches = [TreeNode(label=j, event='S', tstamp=t)
                    for j in range(n)]
        id_counter = n
        
        while branches:
            birth_i, death_i, rho_i, t_i = episodes[i]
            
            losses_to_add = round(len(branches) / rho_i) - len(branches)
            for j in range(losses_to_add):
                branches.append( TreeNode(label=id_counter,
                                          event='L', tstamp=t) )
            id_counter += losses_to_add
            
            while branches:
                w = np.random.exponential( 1 / ((birth_i + death_i) * 
                                                len(branches)) )
                
                if i+1 < len(episodes) and t + w > episodes[i+1][3]:
                    t = episodes[i+1][3]
                    i += 1
                    break
                
                else:
                    t += w
                    
                    if birth_i > np.random.uniform(low=0.0, 
                                                   high=birth_i+death_i):
                        # speciation event drawn
                        spec_node = TreeNode(label=id_counter, event='S',
                                             tstamp=t)
                        id_counter += 1
                        if len(branches) > 1:
                            k, l = np.random.choice(len(branches), 2,
                                                    replace=False)
                            if k > l:
                                k, l = l, k
                            spec_node.add_child(branches[k])
                            spec_node.add_child(branches[l])
                            branches[k] = spec_node
                            branches.pop(l)
                        else:
                            spec_node.add_child(branches[0])
                            tree = Tree(spec_node)
                            # planted root is not a speciation event
                            spec_node.event = None
                            branches.clear()
                    else:
                        # extinction event drawn
                        branches.append( TreeNode(label=id_counter,
                                                  event='L', tstamp=t) )
                        id_counter += 1
        
        # return tree with the following probability
        if np.random.random() < (1 / birth_i) / birth_inv_sum:
            
            return tree
        
    print(f'Could not return a tree after {max_tries} simulations',
          file=sys.stderr)



# -*- coding: utf-8 -*-

"""Evolve sequences along a tree using specific models for substitution, indels,
and rate heterogeneity.

References
----------
.. [1] Z. Yang.
   Computational molecular evolution.
   Oxford series in ecology and evolution. Oxford University Press, 2006.
   ISBN 978-0-19-856699-1 978-0-19-856702-8.
"""

import numpy as np

from asymmetree.seqevolve.EvolvingSequence import EvoSeq, State
from asymmetree.seqevolve.Alignment import AlignmentBuilder
from asymmetree.file_io.SeqFileIO import write_alignment, write_fasta


__author__ = 'David Schaller'


class Evolver:
    
    def __init__(self, subst_model,
                 indel_model=None, het_model=None,
                 gillespie='auto',
                 **kwargs):
        """
        Parameters
        ----------
        subst_model : SubstModel
            Substitution model.
        indel_model : IndelModel, optional
            Model for insertions and deletions, default is None.
        het_model : HetModel, optional
            Model for substitution rate heterogeneity, default is None.
        gillespie : bool, optional
            If True, the Gillespie algorithm is used instead of constructing
            the exchange probability matrix P via matrix diagonalization. The
            Gillespie algorithm is expected to be faster if the rate
            heterogeneity is large. The default is 'auto', in which case the 
            exchange probability matrix is used except if rate heterogeneity
            is enabled.
        """
        
        self.subst_model = subst_model
        
        self.indel_model = indel_model
        self.het_model = het_model
        
        if gillespie == 'auto':
            if het_model and (het_model.classes > 1 or het_model.sitewise):
                self._gillespie = True
            else:
                self._gillespie = False
        else:
            self._gillespie = bool(gillespie)
        
    
    def evolve_along_tree(self, tree, start_length=200, start_seq=None):
        """Evolve sequence along a tree.
        
        If 'start_seq' is not supplied, then the sequence at the root is
        constructed randomly using the stationary probabilities of the
        specified substitution model.
        
        Parameters
        ----------
        tree : Tree
            The tree along which sequences are simulated.
        start_length : int, optional
            The length of the sequence at the root. The default is 200.
            This parameter is ignored when 'start_seq' is supplied.
        start_seq : str, optional
            The start sequence for the root of the tree. The default is None,
            in which case a sequence is constructed at random.
        
        Returns
        -------
        dict
            A dict containing the TreeNode instances in the tree as keys and 
            the simulated sequences as values (instances of type EvoSeq).
        """
        
        self.tree = tree
        self.site_counter = 0
        self.sequences = {}
        
        if start_seq is None:
            root_seq = self._random_sequence(start_length)
        else:
            root_seq = self._initialize_with_sequence(start_seq)
            
        if self.het_model:
            self.het_model.assign(root_seq)
        
        for v in tree.preorder():
            
            if v.parent is None:
                self.sequences[v] = root_seq
            else:
                self.sequences[v] = self._evolve(self.sequences[v.parent],
                                                 v.dist)
                
        return self.sequences
          
            
    def _random_positions(self, n):
        
        return np.random.choice(len(self.subst_model.alphabet), n,
                                p=self.subst_model.freqs)
    
    
    def _random_sequence(self, n):
        
        seq = EvoSeq()
        
        for x in self._random_positions(n):
            seq.append(x, State.ROOT, self.site_counter)
            self.site_counter += 1
        
        return seq
    
    
    def _initialize_with_sequence(self, sequence):
        
        seq = EvoSeq()
        
        for x in self.subst_model.to_indices(sequence):
            seq.append(x, State.ROOT, self.site_counter)
            self.site_counter += 1
        
        return seq
    
    
    def _evolve(self, parent_seq, distance):
        
        child_seq = parent_seq.clone()
        
        if self.indel_model:
            self._generate_indels(child_seq, distance)
            
            if self.het_model:
                self.het_model.assign(child_seq)
        
        if not self._gillespie:
            # use transition prob. matrix
            self._substitute(child_seq, distance)
        else:
            # use Gillespie algorithm
            self._substitute_gillespie(child_seq, distance)
        
        return child_seq
    
    
    def get_sequences(self, include_inner=True):
        """Return the simulated sequence as strings.
        
        Parameters
        ----------
        include_inner : bool, optional
            If True, the sequences of the inner nodes of the tree are also
            includes. The default is True.
        
        Returns
        -------
        dict
            A dictionary with the tree nodes as keys and the str sequence
            (without gaps) as values.
        """
        
        if include_inner:
            return {v: self.subst_model.to_sequence(seq)
                    for v, seq in self.sequences.items()}
        else:
            return {v: self.subst_model.to_sequence(seq)
                    for v, seq in self.sequences.items() if v.is_leaf()}
    
    
    def write_sequences(self, filename, include_inner=True):
        """Write the simulated sequences into a file in fasta format.
        
        Parameters
        ----------
        filename: str
            The path and filename to the file into which the sequences shall
            be written.
        include_inner : bool, optional
            If True, the sequences of the inner nodes of the tree are also
            written to the file. The default is True.
        """
        
        write_fasta(filename, self.get_sequences(include_inner=include_inner))
    
    
    def true_alignment(self, include_inner=True, write_to=None,
                       alignment_format='phylip'):
        """Construct the 'true' alignment of the simulated sequences.
        
        Parameters
        ----------
        include_inner : bool, optional
            If True, the alignment also contains the sequences of the inner
            nodes of the tree. The default is True.
        write_to : str, optional
            The path and filename to the file into which the alignment shall
            be written. The default is None, in which case nothing is written
            to a file.
        alignment_format : str
            Format of the alignment that is written to a file. The default is
            'phylip'.
        
        Returns
        -------
        dict
            A dict containing the TreeNode instances in the tree as keys and 
            the str sequences as values that include the necessary gaps.
        """
        
        alg_builder = AlignmentBuilder(self.tree, self.sequences,
                                       self.subst_model.alphabet,
                                       include_inner=include_inner)
        
        alignment = alg_builder.build()
        
        if write_to:
            write_alignment(write_to, alignment,
                            alignment_format=alignment_format)
        
        return alignment
    
    
    # --------------------------------------------------------------------------
    #               Substitution by transition probability matrix
    # --------------------------------------------------------------------------
    
    def _substitute(self, sequence, t):
        
        # (cumulative) transition probability matrices
        P = {}
        
        r = np.random.random(len(sequence))
        
        for i, site in enumerate(sequence):

            if site.status == State.INHERITED:
                # mutate according to matrix of the corresponding class
                
                d = site.rate_factor * t
                
                if d not in P:
                    if d > 0.0:
                        P[d] = np.cumsum(
                            self.subst_model.transition_prob_matrix(d),
                            axis=1)
                    else:
                        continue
                
                site._value = np.argmax(P[d][site._value, :] > r[i])
            
            else:
                # choose random character
                site._value = np.argmax(self.subst_model.freqs_cumulative > r[i])
        
    
    # --------------------------------------------------------------------------
    #                   Substitution by Gillespie algorithm
    # --------------------------------------------------------------------------
    
    def _substitute_gillespie(self, sequence, t):
        
        for site in sequence:
            
            # choose random characters for insertions
            if site.status == State.INSERTION:
                site._value = np.argmax(self.subst_model.freqs_cumulative 
                                        > np.random.random())
                
            # Gillespie algorithm for inherited positions
            else:
                current_time = 0.0
                while current_time < t:
                    
                    # use negative value on the diagonal of the rate matrix Q
                    rate = - site.rate_factor * self.subst_model.Q[site._value,
                                                                   site._value]
                    
                    current_time += np.random.exponential(1/rate) \
                                    if rate > 0.0 else float('inf')
                    
                    if current_time >= t:
                        break
                        
                    r_mutation = - np.random.random() * self.subst_model.Q[
                                                            site._value,
                                                            site._value]
                    current_sum = 0.0
                    for i in range(len(self.subst_model.alphabet)):
                        
                        # skip the entry on diagonal
                        if i != site._value:
                            current_sum += self.subst_model.Q[site._value, i]
                            
                            if current_sum > r_mutation:
                                mutation = i
                                break
                    
                    site._value = mutation
    
    
    # --------------------------------------------------------------------------
    #                               Indels
    # --------------------------------------------------------------------------
        
    def _generate_indels(self, sequence, t):
        """Generates indels using the Gillespie algorithm."""
        
        current_time = 0.0
        
        while current_time < t:
            
            ins_rate, del_rate = self.indel_model.get_rates(len(sequence))
            total_rate = ins_rate + del_rate
            
            current_time += np.random.exponential(1/total_rate) \
                            if total_rate > 0.0 else float('inf')
            
            if current_time < t:
                
                if np.random.random() < ins_rate / total_rate:
                    self._insertion(sequence)
                else:
                    self._deletion(sequence)
    
    
    def _insertion(self, sequence):
        
        d = self.indel_model.draw_length()
        if d == 0:
            return
        
        pos = np.random.randint(-1, high=len(sequence))
        
        # initialize insertion before the first item
        if pos == -1:
            current_site = sequence.append_left(None, State.INSERTION,
                                                self.site_counter)
            self.site_counter += 1
            d -= 1
            
        # go to the element at position pos
        else:
            current_site = sequence.node_at(pos)
            
        for _ in range(d):
            current_site = sequence.insert_right_of(current_site,
                                                    None, State.INSERTION,
                                                    self.site_counter)
            
            self.site_counter += 1
    
    
    def _deletion(self, sequence):
        
        d = self.indel_model.draw_length()
        if d == 0:
            return
        
        pos = np.random.randint(-d + 1, high=len(sequence))
        
        # deletion begins before or at the start of the sequence
        if pos <= 0:
            d += pos
            pos = 0
            
        sequence.remove_range(pos, d)



# -*- coding: utf-8 -*-

"""
Simulation of noisy distance matrices.

Distances derived from (real-life) gene or protein sequences are burdened with
noise. Such data can either be modeled by simulating sequences, or by
disturbing the distances specified by a given tree directly. This module
implements functions for the latter alternative.
"""

import random
import numpy as np

from tralda.datastructures.Tree import Tree, TreeNode
from asymmetree.tools.PhyloTreeTools import distance_matrix


__author__ = "David Schaller"


# --------------------------------------------------------------------------
#                      RANDOM PERTURBATION NOISE
#   
# --------------------------------------------------------------------------

def noisy_matrix(orig_matrix, sd, metric_repair='reject'):
    """Disturb a distance matrix (which must be a metric) with random noise.
    
    Parameters
    ----------
    orig_matrix : 2-dimensional numpy array
        The distance matrix.
    sd : float
        Disturbance parameter. Serves as standard deviation of a normal
        distrubition with mean 1.0.
    metric_repair : str, optional
        Strategy to ensure that the resulting matrix is still a metric and,
        in particular, satisfies the triangle inequality.
        The default is 'reject', see [1]. Other avaible options are 'general'
        and 'DOMR', see [2].
    
    Returns
    -------
    2-dim. numpy array
        The disturbed distance matrix.
        
    References
    ----------
    .. [1] P. F. Stadler, M. Geiß, D. Schaller, A. López Sánchez, M. González 
       Laffitte, D. Valdivia, M. Hellmuth, M. Hernández Rosales.
       From pairs of most similar sequences to phylogenetic best matches.
       In: Alg. Mol. Biol., 2020, 15, 5.
       doi: 10.1186/s13015-020-00165-2.
    .. [2] A. C. Gilbert, L. Jain.
       If it ain't broke, don't  x it: Sparse metric repair. 
       In: 2017 55th Annual Allerton Conference on Communication, Control, and
       Computing (Allerton), pages 612-619, Monticello, IL, USA, October 2017.
       IEEE. ISBN 978-1-5386-3266-6.
       doi: 10.1109/ALLERTON.2017.8262793.
    """
    
    if metric_repair == 'general':
        return _noise_general_metric_repair(orig_matrix, sd)
    elif metric_repair == 'DOMR':
        return _noise_metric_repair_DOMR(orig_matrix, sd)
    elif metric_repair == 'reject':
        return _noise_reject_method(orig_matrix, sd)
    else:
        raise ValueError("illegal argument '{}'".format(metric_repair))

    
def _noise_reject_method(orig_matrix, sd):
    """Return a matrix D' with noise by accept/reject algorithm."""
    
    D = np.array(orig_matrix, copy=True)
    N = D.shape[0]
    
    success_count = 0
    stop_at = N * (N-1) / 2
    
    while success_count < stop_at:
        i, j = np.random.randint(N, size=2)
        while i == j:
            j = np.random.randint(N)
        old_distance = D[i,j].copy()
        new_distance = 0.0
        while new_distance <= 0.0:
            new_distance = D[i,j] * np.random.normal(loc=1.0, scale=sd)
        D[i,j] = new_distance
        D[j,i] = new_distance
        
        
        if (D[i,j] > np.min(D[i,:] + D[:,j]) or
            np.any( D[:,i] > D[:,j] + D[j,i] ) or
            np.any( D[:,j] > D[:,i] + D[i,j] )):
            D[i,j] = old_distance
            D[j,i] = old_distance
        else:
            success_count += 1
    
    return D


def _noise_metric_repair_DOMR(orig_matrix, sd):
    """Return a matrix D' with noise by metric repair (DOMR) algorithm."""
    
    D = np.array(orig_matrix, copy=True)
    N = D.shape[0]
    
    for i in range(N-1):                        # noise introduction
        for j in range(i+1, N):
            new_distance = 0.0
            while new_distance <= 0.0:
                new_distance = D[i,j] * np.random.normal(loc=1.0, scale=sd)
            D[i,j] = new_distance
            D[j,i] = new_distance
    
    for k in range(N):                          # metric repair: decrease
        for i in range(N):                      # only metric repair (DOMR)
            for j in range(i):                  # with Floyd-Warshall
                if D[i,j] >= D[i,k] + D[k,j]:
                    D[i,j] = D[i,k] + D[k,j]
                    D[j,i] = D[i,k] + D[k,j]
    
    return D


def _noise_general_metric_repair(orig_matrix, sd):
    """Return a matrix D' with noise by metric repair (DOMR) algorithm."""
    
    D = np.array(orig_matrix, copy=True)
    N = D.shape[0]
    
    for i in range(N-1):                        # noise introduction
        for j in range(i+1,N):
            new_distance = 0.0
            while new_distance <= 0.0:
                new_distance = D[i,j] * np.random.normal(loc=1.0, scale=sd)
            D[i,j] = new_distance
            D[j,i] = new_distance
            
    l, r = np.zeros_like(D), np.zeros_like(D)
    for k in range(N):                          # metric repair: general metric repair
        for i in range(N):
            for j in range(i):
                if D[i,j] >= D[i,k] + D[k,j]:
                    l[i,j] += 1
                    r[i,k] += 1
                    r[j,k] += 1
    for k in range(N):                          # metric repair: general metric repair
        for i in range(N):
            for j in range(i):
                if D[i,j] >= D[i,k] + D[k,j]:
                    if l[i,j] > max(r[i,k], r[j,k]):
                        D[i,j] = D[i,k] + D[k,j]
                        D[j,i] = D[i,k] + D[k,j]
                    elif r[i,k] > r[j,k]:
                        D[i,k] = D[i,j] - D[j,k]
                        D[k,i] = D[i,j] - D[j,k]
                    else:
                        D[j,k] = D[i,j] - D[i,k]
                        D[k,j] = D[i,j] - D[i,k]
    
    return D


# --------------------------------------------------------------------------
#                        WRONG TOPOLOGY NOISE
#   
# --------------------------------------------------------------------------
    
def convex_linear_comb(D1, D2, alpha=0.05, first_only=False):
    """Convex linear combination of distance matrices.
    
    Returns the convex linear combinations of two distance matrices
        D1' = (1-alpha) * D1 + alpha * D2
        D2' = (1-alpha) * D2 + alpha * D1
    
    Parameters
    ----------
    D1 : 2-dimensional numpy array
        The first distance matrix.
    D2 : 2-dimensional numpy array
        The second distance matrix.
    alpha : float, optional
        Disturbance parameter, the default is 0.05.
    first_only : bool, optional
        If True, only return the disturbed first matrix. The default is False.
    
    Returns
    -------
    2-dim. numpy array or tuple of 2 matrices
        The disturbed distance matrix or matrices.
        
    References
    ----------
    .. [1] P. F. Stadler, M. Geiß, D. Schaller, A. López Sánchez, M. González 
       Laffitte, D. Valdivia, M. Hellmuth, M. Hernández Rosales.
       From pairs of most similar sequences to phylogenetic best matches.
       In: Alg. Mol. Biol., 2020, 15, 5.
       doi: 10.1186/s13015-020-00165-2.
    """
    if not first_only:
        if D1.shape == D2.shape:
            return ((1-alpha) * D1 + alpha * D2,
                    (1-alpha) * D2 + alpha * D1)
        
        if D1.shape > D2.shape:
            D1, D2 = D2, D1                         # now D1 is smaller
            swap = True
        else:
            swap = False
        
        N1, N2 = D1.shape[0], D2.shape[0]
        D1_alpha = (1-alpha) * D1 + alpha * D2[:N1,:N1]
        D2_alpha = (1-alpha) * D2
        
        for i in range(N2 // N1):
            for j in range(N2 // N1):
                D2_alpha[i*N1:i*N1+N1, j*N1:j*N1+N1] += alpha * D1
                if i == (N2 // N1) - 1 and i == j:
                    D2_alpha[i*N1+N1:N2, j*N1+N1:N2] += alpha * D1[:(N2%N1), :(N2%N1)]
                if i == (N2 // N1) - 1:
                    D2_alpha[i*N1+N1:N2, j*N1:j*N1+N1] += alpha * D1[:(N2%N1), :]
                if j == (N2 // N1) - 1:
                    D2_alpha[i*N1:i*N1+N1, j*N1+N1:N2] += alpha * D1[:, :(N2%N1)]
        
        if swap:
            return D2_alpha, D1_alpha
        else:
            return D1_alpha, D2_alpha
    
    else:
        if D1.shape == D2.shape:
            return (1-alpha) * D1 + alpha * D2
        
        N1, N2 = D1.shape[0], D2.shape[0]
        if N1 < N2:
            return (1-alpha) * D1 + alpha * D2[:N1,:N1]
        else:
            D1_alpha = (1-alpha) * D1
            for i in range(N1 // N2):
                for j in range(N1 // N2):
                    D1_alpha[i*N2:i*N2+N2, j*N2:j*N2+N2] += alpha * D2
                    if i == (N1 // N2) - 1 and i == j:
                        D1_alpha[i*N2+N2:N1, j*N2+N2:N1] += alpha * D2[:(N1%N2), :(N1%N2)]
                    if i == (N1 // N2) - 1:
                        D1_alpha[i*N2+N2:N1, j*N2:j*N2+N2] += alpha * D2[:(N1%N2), :]
                    if j == (N1 // N2) - 1:
                        D1_alpha[i*N2:i*N2+N2, j*N2+N2:N1] += alpha * D2[:, :(N1%N2)]
            return D1_alpha


def wrong_topology_matrix(PGT):
    """Return a wrong topology matrix by rearranging the edges of a binary tree."""
    
    distances = [v.dist for v in PGT.preorder()][1:]    # do not include root,
    if len(distances) % 2 != 0:                         # pruned gene tree (PGT)
        print("List of distances is not even!")         # is binary and not planted,
        return                                          # hence |E| should be even
    random.shuffle(distances)
    
    random_tree = Tree(TreeNode(label=0, dist=0.0))
    id_counter = 1
    current_leaves = [random_tree.root]
    
    while distances:
        v = current_leaves.pop(random.randint(0, len(current_leaves)-1))
        dist1, dist2 = distances.pop(), distances.pop()
        new_child1 = TreeNode(label=id_counter, dist=dist1)
        new_child2 = TreeNode(label=id_counter+1, dist=dist2)
        v.add_child(new_child1)
        v.add_child(new_child2)
        current_leaves.extend(v.children)
        id_counter += 1
    
    random_leaves = [l for l in random_tree.leaves()]   # implicit random bijection
    random.shuffle(random_leaves)                       # to original tree
    
    _, D = distance_matrix(random_tree, leaf_order=random_leaves)
    return D


# --------------------------------------------------------------------------
#                           METRIC CHECK
#   
# --------------------------------------------------------------------------

def _check_metric(matrix):
    """Check whether a given matrix is a metric."""
    
    N = matrix.shape[0]
    for i in range(N):
        if matrix[i,i] != 0.0:
            print("Not all diagonal elements zero!")
            return False
    for i in range(N-1):
        for j in range(i+1,N):
            if matrix[i,j] != matrix[j,i]:
                print("Not symmetrical for",i,j)
                return False
            if abs(matrix[i,j] - np.min(matrix[i,:] + matrix[:,j])) > 1e-8:
                print("Violation of triangle inequality!",
                      i, j, np.min(matrix[i,:] + matrix[:,j]), matrix[i,j])
                return False
    return True


# -*- coding: utf-8 -*-

"""Simulation of species and gene trees.

The subpackage asymmetree.treeevolve contains modules for the simulation and
manipulation of species trees and gene trees. In terms of divergence time,
these trees define an ultrametric on the set of their (extant) leaves. Gene
trees, furthermore, can be manipulated with a realistic rate heterogeneity
among their branches resulting in general additive distances (but no longer
ultrametric).

A typical simulation consists of the following steps:
    
(1) dated species tree (models e.g. 'Yule', and '(episodic) birth-death
    process')
    
(2) dated gene tree(s) (birth-death process with speciations as additional
    branching events)

(3) assignment of asymmetric evolution rates to paralogous genes

(4) pruned gene tree(s) (removal of all branches that lead to losses only)
"""

from asymmetree.treeevolve.SpeciesTree import (species_tree_n,
                                               species_tree_age,
                                               species_tree_n_age,
                                               nonbinary)
from asymmetree.treeevolve.GeneTree import (dated_gene_tree,
                                            GeneTreeSimulator,
                                            prune_losses)
from asymmetree.treeevolve.RateHeterogeneity import (rate_heterogeneity,
                                                     autocorrelation_factors,
                                                     gene_trees)
from asymmetree.treeevolve.DistanceNoise import (noisy_matrix,
                                                 convex_linear_comb,
                                                 wrong_topology_matrix)



# -*- coding: utf-8 -*-

"""Genome simulation.

The module asymmetree.genome.GenomeSimulation provides functions that combine
the simulation of phylogenetic trees and sequences. In particular, the class
GenomeSimulator combines multiple steps described in the previous sections in
order to conveniently simulate whole genomes/proteomes. The (optional) output
directory contains serialized trees, fasta files, and the true alignments.
"""

from asymmetree.genome.GenomeSimulation import GenomeSimulator


# -*- coding: utf-8 -*-

"""
Auxiliary functions for phylogentic trees.

Phylogenetic trees are usually defined as trees where every inner node has
at least two children. However, the class `Tree` does not force this property.
Simulated trees are often 'planted' i.e. the root has only a single child,
and this edge represents the ancestral lineage.
"""

import collections, itertools, random, re

import numpy as np

from tralda.datastructures.Tree import Tree, TreeNode


__author__ = 'David Schaller'


default_attributes = {'label': '',
                      'event': None,
                      'reconc': None,
                      'dist': 1.0,
                      'tstamp': None,
                      'transferred': 0}


def node_to_str(node):
    """String representation of a node in a phylogenetic tree.
    
    Returns
    -------
    str
        String representation of the node including reconc (if defined)
        and dist.
    """
    
    if isinstance(node.reconc, (tuple, list)):
        return '{}<{}-{}>:{}'.format(node.label, *node.reconc, node.dist)
    elif node.reconc:
        return '{}<{}>:{}'.format(node.label, node.reconc, node.dist)
    else:
        return '{}:{}'.format(node.label, node.dist)


def sorted_nodes(tree, oldest_to_youngest=True):
    """List of nodes sorted by timestamp.
    
    Return a list of all nodes in the tree sorted by time stamp beginning
    with the oldest (highest time stamp). Optionally the order can be
    reversed.
    
    Parameters
    ----------
    tree : Tree
        A phylogenetic tree with nodes that have a `tstamp` attribute.
    oldest_to_youngest : bool, optional
        If True, the nodes are sorted from oldest to youngest, otherwise
        from youngest to oldest (the default is True).
        
    Returns
    -------
    list of TreeNode instances
        Nodes sorted by their timestamp.
    """
    
    return sorted(tree.preorder(),
                  key=lambda node: node.tstamp,
                  reverse=oldest_to_youngest)


def sorted_edges(tree):
    """List of edges (u,v) sorted by timestamp of u.
    
    Return a list of all edges (u,v) sorted by timestamp of u beginning
    with the oldest (highest time stamp).
    
    Parameters
    ----------
    tree : Tree
        A phylogenetic tree with nodes that have a `tstamp` attribute.
        
    Returns
    -------
    list of pairs of TreeNode instances
        Edges sorted by their timestamp.
    """

    return sorted(tree.edges(),
                  key=lambda e: (e[0].tstamp, e[1].tstamp),
                  reverse=True)

def distance_from_timing(tree):
    """Adjusts all distances according to the time stamp difference.
    
    Parameters
    ----------
    tree : Tree
        A phylogenetic tree with nodes that have a `tstamp` attribute.
    """
    
    if tree.root:
        tree.root.dist = 0.0
    
    for u, v in tree.edges():
        v.dist = abs(u.tstamp - v.tstamp)


def delete_and_reconnect(tree, node,
                         add_distances=True,
                         keep_transferred=True):
    """Delete a node from the tree and reconnect its parent and children.
    
    Parameters
    ----------
    node : TreeNode
        The node in the tree to be deleted.
    add_distances : bool, optional
        When the node v is deleted and its children are reconnected to its
        parent, add to the `dist` parameter of the children `v.dist`, i.e.
        the distance of v from its parent (the default is True).
    keep_transferred : bool, optional
        When the edge of the deleted node v from its parent u was a
        transfer edge, make all edges from u to the children of v transfer
        edges (the default is True).
        
    Returns
    -------
    TreeNode or bool
        The parent of the node, if it could be deleted, or False, if the
        node could not be deleted, i.e., it has no parent.
    """
    
    if not node.parent:
        return False
    
    for child in node.children:
        if add_distances and hasattr(node, 'dist') and hasattr(child, 'dist'):
            child.dist += node.dist
        if (keep_transferred and 
            hasattr(node, 'transferred') and 
            node.transferred):
            child.transferred = 1
    
    return tree.delete_and_reconnect(node)


def add_planted_root(tree):
    """Add a new root that has the original root as its single child.
    
    Parameters
    ----------
    tree : Tree
        A phylogenetic tree.
    
    Returns
    -------
    TreeNode
        The newly added root.
    """
    
    old_root = tree.root
    tree.root = TreeNode()
    for k, v in default_attributes.items():
        setattr(tree.root, k, v)
    
    if old_root:
        tree.root.add_child(old_root)
    
    return tree.root


def reconc_sorted_leaves(tree, return_list=False):
    """Sort leaves by their reconciliation attribute.
    
    Parameters
    ----------
    tree : Tree
        A tree with nodes that have a `reconc` attribute.
    return_list : bool, optional
        If True, also return a list of leaves such that leaves of the same
        reconciliation appear as a coherent sequence (the default is False).
    
    Returns
    -------
    dict
        A dictionary with reconciliations as keys and a list of corresponding
        nodes as values.
    list, optional
        List of leaves such that leaves of the same reconciliation appear as a
        coherent sequence.
    """
    
    reconc_dict = {}
    
    for leaf in tree.leaves():
        if leaf.reconc not in reconc_dict:
            reconc_dict[leaf.reconc] = []
        reconc_dict[leaf.reconc].append(leaf)
    
    if not return_list:
        return reconc_dict
    else:
        leaves = []
        for reconc, leaf_list in reconc_dict.items():
            for leaf in leaf_list:
                leaves.append(leaf)
        
        return reconc_dict, leaves
    
    
def distance_matrix(tree, leaf_order=None):
    """Distance matrix on the leaf set of the phylogenetic tree.
    
    Computes a distance matrix on the set of leaves of the tree where each
    distances is the sum of the distances (`dist`) on the path connecting
    the pair of leaves.
    
    Additionally a list of leaves corresponding to the indices in the
    matrix is returned.
    
    Parameters
    ----------
    tree : Tree
        A tree with nodes that have a `dist` attribute.
    leaf_order : list, optional
        A list of all leaves in the tree defining the indices for the
        matrix (the default is None, in which case leaves are indexed in
        sibling order).
    
    Returns
    -------
    list of TreeNode objects
        Represents the order for the lines/columns in the distance matrix.
    numpy.ndarray (dtype=numpy.float)
        The distance matrix.
    """
    
    distance_dict = distances_from_root(tree)
    leaves = tree.leaf_dict()
    
    if leaf_order:
        if set(leaf_order) != set(leaves[tree.root]):
            raise ValueError('ordered leaf list does not match with the '\
                             'leaves in the tree')
        L = leaf_order
    else:
        # leaves in sibling order
        L = leaves[tree.root]     
    
    leaf_index = {l: i for i, l in enumerate(L)}
    
    D = np.zeros((len(L),len(L)), dtype=np.float)
    
    for v in tree.preorder():
        if v.children:
            for c1, c2 in itertools.combinations(v.children, 2):
                for x in leaves[c1]:
                    x_index = leaf_index[x]
                    x_dist = distance_dict[x] - distance_dict[v]
                    for y in leaves[c2]:
                        y_index = leaf_index[y]
                        y_dist = distance_dict[y] - distance_dict[v]
                        D[x_index, y_index] = x_dist + y_dist
                        D[y_index, x_index] = x_dist + y_dist
    
    return L, D


def distances_from_root(tree):
    """The distances of each node to the root of the tree.
    
    Parameters
    ----------
    tree : Tree
        A tree with nodes that have a `dist` attribute.
    
    Returns
    -------
    dict
        The keys are TreeNode objects and the values their distances
        (sum of `dist`) to the root.
    """
    
    distance_dict = {}
    
    for v in tree.preorder():
        if not v.parent:
            distance_dict[v] = 0.0
        else:
            depth = distance_dict[v.parent] + v.dist
            distance_dict[v] = depth
            
    return distance_dict


def topology_only(tree, inplace=True):
    """Reset reconciliations, distances, time stamps, transfer status, and
    inner labels.
    
    Parameters
    ----------
    tree : Tree
        A tree.
    inplace : bool
        If True, reset the attributes of this tree instance, otherwise
        make a copy first and modify the copy.
    
    Returns
    -------
    Tree
        The original or a copy of the tree instance with reset attributes.
    """
    
    if not inplace:
        T = tree.copy()
    else:
        T = tree
    
    for v in T.preorder():
        if v.children:
            v.label = ''
            v.reconc = None
        v.dist = 1.0
        v.tstamp = None
        v.transferred = 0
    
    return T


def count_node_types(tree):
    """Count speciations, duplication, losses, HGTs and surviving genes.
    
    Parameters
    ----------
    tree : Tree
        A tree with nodes that have a `label` attribute.
    
    Returns
    -------
    dict
        With the event counts as values. Key are 'S' (speciations), 'D'
        (duplication), 'L' (losses), 'H' (HGTs) and 'extant' (surviving
        genes).
    """
    
    counts = {'S': 0, 'D': 0, 'L': 0, 'H': 0, 'extant': 0}
    
    for v in tree.preorder():
        
        if not v.children:
            if v.event == 'L':
                counts['L'] += 1
            else:
                counts['extant'] += 1
        
        elif v.event == 'S':
            counts['S'] += 1
        elif v.event == 'D':
            counts['D'] += 1
        elif v.event == 'H':
            counts['H'] += 1
    
    return counts


def random_colored_tree(n, colors, binary=False, force_all_colors=False):
     """Create a random colored tree.
     
     The number of leaves and the reconciliation labels are specified in the
     parameters n and colors, respectively. Each non-leaf node in the 
     resulting tree will have at least children (property of phylogenetic
     trees).
     
     Parameters
     ----------
     n : int
         The desired number of leaves.
     colors : int or list
         The list of recociliations, or the desired number in which case the
         reconciliations {1, ..., colors} are used.
     binary : bool, optional
         If True, forces the tree to be binary (the default is False).
     force_all_colors : bool
         If True, the resulting tree is guaranteed to have at least one leaf
         of each reconciliation (the default is False).
     
     Returns
     -------
     Tree
         A random tree with n leaves to which the `reconc` attribute is
         assigned at random.
     
     Raises
     ------
     TypeError
         If n is not an integer > 0.
     ValueError
         If the number of colors is greater than n and `force_all_colors`
         is true.
     """
     
     tree = Tree.random_tree(n, binary=binary)
         
     if isinstance(colors, int):
         colors = [i+1 for i in range(colors)]
     elif not isinstance(colors, collections.abc.Iterable):
         raise TypeError("'colors' must be of type 'int' or iterable")
         
     if len(colors) > n and force_all_colors:
         raise ValueError('cannot force all colors since #colors > n')
     
     leaves = [l for l in tree.leaves()]
     
     if force_all_colors:
         # use every color at least once
         permutation = np.random.permutation(len(leaves))
         for i in range(len(leaves)):
             if i < len(colors):
                 leaves[permutation[i]].reconc = colors[i]
             else:
                 # color the remaining leaves randomly
                 leaves[permutation[i]].reconc = random.choice(colors)
     else:
         # assign colors completely randomly
         for leaf in leaves:
             leaf.reconc = random.choice(colors)
             
     return tree
 

def random_ultrametric_timing(tree, inplace=False, adjust_distances=False):
    """Generate a random ultrametric timing for the tree.
    
    Parameters
    ----------
    tree : Tree
        The tree for which a random timing shall be generated.
    inplace : bool, optional
        If True, the input tree is modified, otherwise a copy is returned.
        The default is False.
    adjust_distances : bool, optional
        If True, also adjust the dist attribute of the tree nodes to match the
        differences of the tstamp values. The default is False.
    
    Returns
    -------
    Tree
        A random tree whose nodes have tstamp attributes that represent the
        generated random ultrametric timing.
    """
    
    if not inplace:
        tree = tree.copy()
        
    for v in tree.preorder():
        if not v.children:
            v.tstamp = 0.0
        elif not v.parent:
            v.tstamp = 1.0
        else:                               # random walk to a leaf
            pos = v                         # current position
            length = 0                      # path length |P|
            while pos.children:
                length += 1
                pos = pos.children[np.random.randint(len(pos.children))]
            v.tstamp = (
                v.parent.tstamp * (1 - 2 * np.random.uniform() / (length+1))
            )
    
    if adjust_distances:
        distance_from_timing(tree)
             
    return tree
 
    
def phylo_tree_attributes(tree, inplace=True):
    """Add the attributes for a phylogentic tree if not already set.
    
    Parameters
    ----------
    tree : Tree
    inplace : bool
        If True, the input tree is modified, otherwise a copy is returned.
        The default is True.
    
    Returns
    -------
    Tree
    """
    
    if not inplace:
        tree = tree.copy()
    
    for v in tree.preorder():
        for key, value in default_attributes.items():
            if not hasattr(v, key):
                setattr(tree.root, key, value)
    
    return tree


# --------------------------------------------------------------------------
#                    RECONSTRUCTION OF INFORMATION
# --------------------------------------------------------------------------

def assign_missing_labels(tree):
    """Assign integer label to nodes with no or missing label.
    
    Also assign the `number_of_species` attribute, i.e., the number of leaves,
    to the tree.
    
    Parameters
    ----------
    tree : Tree
    """
    
    tree.number_of_species = 0
    labels = set()
    
    for v in tree.preorder():
        if not v.children:
            tree.number_of_species += 1
            
        if hasattr(v, 'label') and v.label:
            labels.add(v.label)
    
    # assign new labels to remaining nodes
    current_label = 0
    for v in tree.preorder():
        if not hasattr(v, 'label') or not v.label:
            while current_label in labels:
                current_label += 1
            v.label = current_label
            labels.add(current_label)


def reconstruct_reconc_from_graph(tree, G):
    """Reconstruct the reconciliations from a NetworkX Graph.
    
    Parameters
    ----------
    tree : Tree
    G : networkx.Graph
        The graph from which labels and reconciliations shall be reconstructed.
    """
    
    for v in tree.preorder():
        if hasattr(v, 'label') and v.label in G:
            if 'reconc' in G.nodes[v.label]:
                if isinstance(G.nodes[v.label]['reconc'], list):
                    v.reconc = tuple(G.nodes[v.label]['reconc'])
                else:
                    v.reconc = G.nodes[v.label]['reconc']
            elif 'color' in G.nodes[v.label]:
                if isinstance(G.nodes[v.label]['color'], list):
                    v.reconc = tuple(G.nodes[v.label]['color'])
                else:
                    v.reconc = G.nodes[v.label]['color']
            
            
def reconstruct_timestamps(tree):
    """Reconstruct the timestamps.
    
    Make the time stamps matching with the distance attribute. The root
    obtains time stamp 1.0, and all other node smaller time stamps such
    that the difference to the parent's time stamp is exactly `dist`.
    
    Parameters
    ----------
    tree : Tree
        A tree with nodes that have a `dist` attribute.
    """
    
    tree.root.tstamp = 1.0
    for v in tree.preorder():
        if v.parent:
            v.tstamp = v.parent.tstamp - v.dist
    

# --------------------------------------------------------------------------
#                         TREE MANIPULATION
# -------------------------------------------------------------------------- 

def delete_losses_and_contract(tree, inplace=False):
    """Delete all branches leading to loss leaves only.
    
    Nodes that would have only a single child afterwards are suppressed (except
    possibly the planted root), i.e. their children are recursively reconnected
    to the parents. Distances are cumulated in this process and the transferred
    status is kept in the sense that an edge is a transfer edge if at least
    one edge on the contracted path to which it corresponds was a transfer edge.
    
    Parameters
    ----------
    tree : Tree
        The tree in which loss branches shall be removed.
    inplace : bool, optional
        If True, the tree is directly manipulated. The default is False in
        which case a copy of the tree is created which gets manipulated while
        the original tree remains untouched.
    
    Returns
    -------
    Tree
        The tree with all loss branches removed (original instance or a new one
        depending on the `inplace` parameter).
    """
    
    if not inplace:
        tree = tree.copy()
    
    loss_nodes = []
    for node in tree.postorder():
        if not node.children and node.event == 'L':
            loss_nodes.append(node)
    
    # traverse from loss node to root delete if degree <= 1
    for loss_node in loss_nodes:
        current = delete_and_reconnect(tree, loss_node,
                                       add_distances=True,
                                       keep_transferred=True)
        
        while len(current.children) < 2 and current.parent:
            current = delete_and_reconnect(tree, current,
                                           add_distances=True,
                                           keep_transferred=True)
    
    return tree


def remove_planted_root(tree, inplace=True):
    """Remove the planted root of the tree (if existent).
    
    Parameters
    ----------
    tree : Tree
        The tree in which the planted root shall be removed.
    inplace : bool, optional
        If True, the tree is directly manipulated, otherwise a copy is created
        and the original tree remains untouched (the default is True).
    
    Returns
    -------
    Tree
        The tree with the planted root removed (original instance or a new one
        depending on the `inplace` parameter).
    """
    
    if not inplace:
        tree = tree.copy()
        
    # delete the root if the tree is planted
    if len(tree.root.children) == 1:
        new_root = tree.root.children[0]
        new_root.detach()
        tree.root = new_root
        new_root.dist = 0.0
        
    if not tree.root.children and not tree.root.label:
        # no surviving genes --> return empty tree
        tree.root = None
    
    return tree


# --------------------------------------------------------------------------
#                          TREE  <--->  NEWICK
# --------------------------------------------------------------------------
        
def to_newick(tree, label=True, reconc=True, distance=True,
                    label_inner=True, reconc_inner=False):
    """Return a Newick representation of the tree.
    
    This function overrides the function of the parent class.
    
    Parameters
    ----------
    label : bool, optional
        If True, the Newick str contains the labels of the nodes (the
        default is True).
    reconc : bool, optional
        If True, the Newick str contains the reconciliations of the nodes in
        <[...]> brackets (the default is True).
    distance : bool, optional
        If True, the Newick str contains the distances of the nodes in
        standard :[...] notation (the default is True).
    label_inner : bool, optional
        If True, the Newick str also contains the labels of the inner 
        nodes (the default is True).
    reconc_inner : bool, optional
        If True, the Newick str contains the reconciliations of the inner nodes
        (the default is False).
    
    Returns
    -------
    str
        A Newick representation of the tree.
    """
    
    def _to_newick(node):
        
        if not node.children:
            token = ''
            if label and hasattr(node, 'label'):
                token += str(node.label)
            if reconc and hasattr(node, 'reconc') and node.reconc:
                token += '<{}-{}>'.format(*node.reconc) \
                    if isinstance(node.reconc, (tuple, list)) \
                    else '<{}>'.format(node.reconc)
            if distance and hasattr(node, 'dist'):
                token += ":{}".format(node.dist)
            return token
        else:
            s = ''
            for child in node.children:
                s += _to_newick(child) + ','
            token = ''
            if label and hasattr(node, 'label') and label_inner:
                token += str(node.label)
            if reconc_inner and hasattr(node, 'reconc') and node.reconc:
                token += '<{}-{}>'.format(*node.reconc) \
                    if isinstance(node.reconc, (tuple, list)) \
                    else '<{}>'.format(node.reconc)
            if distance and hasattr(node, 'dist'):
                token += ':{}'.format(node.dist)
            return '({}){}'.format(s[:-1], token)
    
    
    if tree.root:
        return _to_newick(tree.root) + ';'
    else:
        return ';'


def parse_newick(newick):
    """Parses trees in Newick format into object of type 'Tree'.
    
    Parameters
    ----------
    newick : str
        A tree in Newick format.
    
    Returns
    -------
    Tree
        The parsed tree.
    
    Raises
    ------
    TypeError
        If the input is not a string.
    ValueError
        If the input is not a valid Newick string.
    
    Notes
    -----
    Do not use this function for serialization and reloading Tree
    objects. Use the `serialize()` function instead.
    Labels and reconciliations that are integer numbers are converted to int.
    The reconciliations (if present in <...> in the string) are parsed as
    strings and need to be converted to integers afterwards if necessary.
    """
    
    def to_int(item):
        """Trys to convert the string into int."""
        
        return int(item) if item.isdigit() else item
    
    tree = Tree.parse_newick(newick)
    
    # regex for notation label<reconc>
    label_col_regex = re.compile(r"'?([a-zA-Z0-9_]*)'?<(.*)>")
    
    for node in tree.preorder():
        
        node.event = ''
        
        if not hasattr(node, 'dist'):
            node.dist = 1.0
        
        label = str(node.label)
        
        label_col = label_col_regex.match(label)
        if label_col:
            node.label = to_int(label_col.group(1))
            node.reconc = to_int(label_col.group(2))
        else:
            node.reconc = None
        
        # reconc is a tuple
        if (node.reconc and 
            isinstance(node.reconc, str) and 
            node.reconc.find('-') != -1):
            
            a, b = node.reconc.split('-')
            node.reconc = (to_int(a), to_int(b))
    
    return tree



# -*- coding: utf-8 -*-

"""
Evolution rate heterogeneity.

Introduce evolution rate asymmetries and autocorrelation.
"""

from warnings import warn

import numpy as np

from asymmetree.treeevolve.GeneTree import GeneTreeSimulator
from asymmetree.tools.Sampling import Sampler
from asymmetree.tools.PhyloTreeTools import sorted_nodes


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                         USER INTERFACE FUNCTION
# --------------------------------------------------------------------------

def gene_trees(S,
               n=1,
               dupl_rate=0.0,
               loss_rate=0.0,
               hgt_rate=0.0,
               base_rate=1.0,
               **kwargs):
    """Simulates dated gene trees with non-ultrametric edge lengths along a
    species tree.
    
    Parameters
    ----------
    S : Tree
        The species tree along which the gene trees are simulated.
    n : int, optional
        Number of gene trees to be simulated, default is 1, in which case a
        tree is returned, otherwise a list is returned.
    dupl_rate : float or tuple
        The (distribution for the) duplication rate, see documentation for
        available option. The default is constant 0.0.
    loss_rate : float or tuple
        The (distribution for the) loss rate, see documentation for
        available option. The default is constant 0.0.
    hgt_rate : float or tuple
        The (distribution for the) HGT rate, see documentation for
        available option. The default is constant 0.0.
    base_rate : float or tuple
        The (distribution for the) evolution rate at the roots of the gene
        trees, see documentation for available options. The default is
        constant 1.0.
    kwargs : optional
        See documentation or parameters of GeneTreeSimulator.simulate() and
        rate_heterogeneity() for additional parameters.
    
    Returns
    -------
    list
        A list of simulated gene tree.
    """
    
    gene_trees = []
    simulator = GeneTreeSimulator(S)
    
    dupl_rate_sampler = Sampler(dupl_rate)
    loss_rate_sampler = Sampler(loss_rate)
    hgt_rate_sampler = Sampler(hgt_rate)
    base_rate_sampler = Sampler(base_rate)
    
    # autocorrelation between genes of the same or related species
    autocorr_variance = kwargs.pop('autocorr_variance', 0.0)
    _, autocorr_factors = autocorrelation_factors(S, autocorr_variance)
    
    # main simulation and imbalancing
    for i in range(n):
        TGT = simulator.simulate(dupl_rate=dupl_rate_sampler(),
                                 loss_rate=loss_rate_sampler(),
                                 hgt_rate=hgt_rate_sampler(),
                                 **kwargs)
        rate_heterogeneity(TGT, S,
                           base_rate=base_rate_sampler(),
                           autocorr_factors=autocorr_factors,
                           **kwargs)
        gene_trees.append(TGT)
    
    return gene_trees
    

# --------------------------------------------------------------------------
#                      evolution rate heterogeneity
# --------------------------------------------------------------------------

def rate_heterogeneity(T, S,
                       base_rate=1.0,
                       autocorr_factors=None,
                       autocorr_variance=0.0,
                       rate_increase=('gamma', 0.5, 2.2),
                       CSN_weights=(1, 1, 1),
                       inplace=True,
                       **kwargs):
    """Introduces evolution rate heterogeneity into a gene tree.
    
    The function applies rate multiplier for three sources of evolution rate
    heterogeneity: (i) gene-family-specific heterogeneity is modeled through
    the 'base_rate' parameter; (ii) for species-specific heterogeneity an
    autocorrelated relaxed molecular clock model as in [1] is used; and (iii)
    for paralog-secific heterogeneity it is chosen between the three modes
    conservation, subfunctionalization, and neofunctionalization for the rates
    of the descendant lineages of a duplication event.
    The assigned rates are used to modify the length ('dist') of the edges of
    the (originally ultrametric) dated gene tree.
    
    Parameters
    ----------
    T : Tree
        The gene tree.
    S : Tree
        The species tree.
    base_rate : float, optional
        Mean of substitution rate for conserved genes.
    autocorr_factors : dict, optional
        A dictonary containing autocorrelation rate factors for the edges of S
        (key = v.label for edge (v.parent, v); value = the rate as a float).
        The default is None, in which case autocorrelation factors are 
        generated if 'autocorr_variance' > 0.0, or no such modification is
        applied. See [1] for theoretical background.
    autocorr_variance : float, optional
        Autocorrelation variance factor for a lognormal distribution, only
        considered if 'autocorrelation_rates' are not supplied. See [1] for
        theoretical background.
    rate_increase : float or tuple, optional
        Distribution of the (relative) rate increase (w.r.t. the base rate)
        for divergent genes, i.e. to a factor 1 + x. The default is a Gamma
        distribution with shape 0.5 and scale 2.2, which was fitted to the
        data in [2].
    CSN_weights : tuple, optional
        Weights for choice between conservation, subfunctionalization and
        neofunctionalization. The default is (1, 1, ,1), i.e., all three modes
        are equally likely to be chosen at each duplication event.
    inplace : bool, optional
        If False, copy the tree before imbalancing. The deafault is True.
    
    Returns
    -------
    Tree
        The original instance of the gene tree (inplace=True) or a copy of
        the gene tree (inplace=False) with modified 'dist' attributes of the
        nodes.
    
    References
    ----------
    .. [1] H. Kishino, J. L. Thorne, and W. J. Bruno.
       Performance of a Divergence Time Estimation Method under a Probabilistic
       Model of Rate Evolution. 
       In: Molecular Biology and Evolution, 18(3):352-361, March 2001.
       doi: 10.1093/oxfordjournals.molbev.a003811.
    .. [2] K. P. Byrne and K. H. Wolfe.
       Consistent Patterns of Rate Asymmetry and Gene Loss Indicate Widespread
       Neofunctionalization of Yeast Genes After Whole-Genome Duplication.
       In: Genetics, 175(3):1341-1350, March 2007.
       doi: 10.1534/genetics.106.066951.
    """
    
    if not inplace:
        T = T.copy()
     
    # paralog-specific rate heterogeneity (neo- and subfunctionalization)
    _divergent_rates(T, S, 
                     Sampler(rate_increase, shift=1.0), 
                     np.asarray(CSN_weights) / sum(CSN_weights))
    
    # species-specific rate heterogeneity (autocorrelation)
    if not autocorr_factors and autocorr_variance > 0.0:
        _, autocorr_factors = autocorrelation_factors(S, autocorr_variance)
    if autocorr_factors:
        for v in T.preorder():
            if v.parent:
                edge_ID = v.reconc[1] if isinstance(v.reconc, (tuple, list)) \
                                     else v.reconc
                v.dist *= autocorr_factors[edge_ID]
    
    # gene-family-specific rate heterogeneity (base rate)
    for v in T.preorder():
        v.dist *= base_rate
    
    return T

def assign_rates(T, S, **kwargs):
    
    warn('This method is deprecated. Use rate_heterogeneity() instead.',
         DeprecationWarning, stacklevel=2)
    return rate_heterogeneity(T, S, **kwargs)


# --------------------------------------------------------------------------
#     paralog-specific rate heterogeneity (neo- and subfunctionalization)
# --------------------------------------------------------------------------

def _adjust_distances(T, rates):
    
    for edge, rate_list in rates.items():
        time_points = np.asarray([tstamp for tstamp, _ in rate_list] +
                                 [edge[1].tstamp])
        rate_values = np.asarray([rate for _, rate in rate_list])
        edge[1].dist = np.dot(-np.diff(time_points), rate_values)


def _divergent_rates(T, S, sampler, CSN_weights):
    """Assign divergent genes and manipulate the distances in the gene tree.
    
    Parameters
    ----------
    T : Tree
        The gene tree.
    S : Tree
        The species tree.
    sampler : asymmetree.tools.Sampling.Sampler
        Sampler for rate increase for divergent genes.
    CSN_weights : tuple
         Weights for choice between conservation, subfunctionalization and
         neofunctionalization.
    
    Returns
    -------
    Tree
        The original gene tree instance with manipulated 'dist' attributes of
        its nodes.
    """
    
    T_nodes = sorted_nodes(T)
    
    # edge --> list of (tstamp, rate) tuples
    rates = {edge: [] for edge in T.edges()}
    
    S_parents = {v.label: v.parent.label for v in S.preorder() if v.parent}
    gene_counter = {(u.label, v.label): [] for u, v in S.edges()}
    
    # nodes willl be marked as conserved or divergent
    marked = {v: 'conserved' for v in T_nodes}
    
    for u in T_nodes:
        
        # ----------------- SPECIATION -----------------
        if u.event in ('S', '', None):
            for v in u.children:
                marked[v] = marked[u]
                S_u = u.reconc
                S_v = v.reconc if not isinstance(v.reconc, (tuple, list)) \
                      else v.reconc[1]
                gene_counter[(S_u, S_v)].append(v)
                new_rate = sampler() if marked[v] == 'divergent' else 1.0
                rates[(u,v)].append((u.tstamp, new_rate))
            
        # ---------------- DUPLICATION -----------------
        elif u.event == 'D' or u.event == 'GC':
            
            gene_counter[u.reconc].remove(u)
            
            # will contain the indices of the conserved descendants
            conserved = set()
            
            if marked[u] != 'divergent':
                # if parental lineage was divergent or subfunctionalization
                # is drawn below, all descendant lineages are divergent
                
                r = np.random.choice(3, p=CSN_weights)
                if r == 0:
                    # conservation: all descendants conserved
                    conserved.update(range(len(u.children)))
                elif r == 2:
                    # neofunctionalization: exactly 1 descendant conserved
                    conserved.add(np.random.randint(len(u.children)))
                    
            for i, v in enumerate(u.children):
                
                if i in conserved:
                    marked[v] = 'conserved'
                    new_rate = 1.0
                else:
                    marked[v] = 'divergent'
                    new_rate = sampler()
                    
                gene_counter[u.reconc].append(v)
                rates[(u,v)].append((u.tstamp, new_rate))
        
        # ------------------- LOSS ---------------------
        elif u.event == 'L':
            gene_counter[u.reconc].remove(u)
            
            if len(gene_counter[u.reconc]) == 1:
                # if only one lineage remains in the species, set its status
                # to conserved
                v = gene_counter[u.reconc][0]
                if marked[v] == 'divergent':
                    marked[v] = 'conserved'
                    rates[(v.parent,v)].append((u.tstamp, 1.0))
        
        # ---------- HORIZONTAL GENE TRANSFER ----------
        elif u.event == 'H':
            v1, v2 = u.children
            if v1.transferred:
                v1, v2 = v2, v1         # now v2 is the transferred copy
                
            # untransferred copy
            marked[v1] = marked[u]
            gene_counter[u.reconc].remove(u)
            gene_counter[u.reconc].append(v1)
            if u.parent:
                rates[(u,v1)].append((u.tstamp, rates[(u.parent,u)][-1][1]))
            else:
                new_rate = sampler() if marked[v1] == 'divergent' else 1.0
                rates[(u,v1)].append((u.tstamp, new_rate))
            
            # transferred copy
            S_edge = v2.reconc if isinstance(v2.reconc, (tuple, list)) \
                     else (S_parents[v2.reconc], v2.reconc)
            gene_counter[S_edge].append(v2)
            if len(gene_counter[S_edge]) == 1:
                marked[v2] = 'conserved'
                new_rate = 1.0
            else:
                marked[v2] = 'divergent'
                new_rate = sampler()
            rates[(u,v2)].append((u.tstamp, new_rate))
            
    _adjust_distances(T, rates)
    
    return T

# --------------------------------------------------------------------------
#           species-specific rate heterogeneity (autocorrelation)
# --------------------------------------------------------------------------
    
def autocorrelation_factors(tree, variance):
    """Lognormal model to assign rate factors to a species tree.
    
    The parameter 'variance' is a hyperparameter for a log-normal distribution
    from which offspring rates are drawn. The overall variance of this
    distribution is 'variance' * divergence time.
    The rates are first computed for the nodes, the rates of the edges are
    assigned afterwards as the arithmetic mean of the rates of the two incident
    nodes.
    
    Parameters
    ----------
    tree : Tree
        The species tree.
    variance : float
        The hyperparameter for a log-normal distribution from which offspring
        rates are drawn at each node.
    
    Returns
    -------
    tuple of two dicts
        A dict mapping the labels of the nodel to their assigned rated, and a
        second dict mapping the labels of v of edges (v.parent, v) to the
        assigned rates of the edges.
    
    References
    ----------
    .. [1] H. Kishino, J. L. Thorne, and W. J. Bruno.
       Performance of a Divergence Time Estimation Method under a Probabilistic
       Model of Rate Evolution. 
       In: Molecular Biology and Evolution, 18(3):352-361, March 2001.
       doi: 10.1093/oxfordjournals.molbev.a003811.
    """
    
    node_rates = {}                 # maps node v --> rate of v
    edge_rates = {}                 # maps v of edge (u,v) --> rate of (u,v)
    
    for v in tree.preorder():
        if not v.parent:
            # assign factor 1.0 to root (= expected value for all other nodes
            # and edges)
            node_rates[v.label] = 1.0
            edge_rates[v.label] = 1.0
        else:
            var = variance * v.dist
            # ensure that exp. value is equal to parent's rate
            mu = np.log(node_rates[v.parent.label]) - var/2
            
            node_rates[v.label] = np.exp(np.random.normal(mu, np.sqrt(var)))
            
            # edge rate as arithmetic mean of u and v
            edge_rates[v.label] = (node_rates[v.parent.label] + 
                                   node_rates[v.label]) / 2
            
    return node_rates, edge_rates



# -*- coding: utf-8 -*-

"""Calculation of sequence distances for various models.

References
----------
.. [1] Z. Yang.
   Computational molecular evolution.
   Oxford series in ecology and evolution. Oxford University Press, 2006.
   ISBN 978-0-19-856699-1 978-0-19-856702-8.
"""

import numpy as np
import scipy.optimize

from asymmetree.seqevolve import SubstModel


__author__ = 'David Schaller'


MIN_LENGTH = 1e-6
MAX_LENGTH = 20.0
#MAXITER = 100


def maximum_likelihood_distance(seq1, seq2,
                                subst_model=None,
                                model_type=None, model_name=None):
    
    if subst_model is None:
        
        if model_type is None or model_name is None:
            raise ValueError('no substitution model specified')
        else:
            subst_model = SubstModel(model_type, model_name)
            
    seqs = _to_indices(seq1, seq2, subst_model)
    
    x0 = _initial_guess(seqs, subst_model.model_type)
    
    # distance 0.0 if sequences are equal
    if x0 == 0.0:
        return x0
    
    opt_result = scipy.optimize.minimize(_likelihood, x0, 
                                         args=(seqs, subst_model),
                                         bounds=((MIN_LENGTH, None),),)
    
    if opt_result.success:
        return opt_result.x
    else:
        return float('nan')
    

def _initial_guess(seqs, model_type):
    
    p = np.sum(seqs[0] != seqs[1]) / seqs.shape[1]
    
    x0 = _JC69_transform(p, amino_acid=(model_type=='a'))
    
    x0 = x0 if x0 != float('nan') else MAX_LENGTH / 2
    
    return x0
    

def _likelihood(x, seqs, subst_model):
    
    P = subst_model.transition_prob_matrix(x)
    
    # - log-likelihood
    return - np.sum( np.log( P[seqs[0], seqs[1]] ))
            
            
def _to_indices(seq1, seq2, subst_model):
    
    if len(seq1) != len(seq2):
        raise ValueError('unequal sequence lengths: {len(seq1)} and {len(seq2)}')

    seq1_indeces, seq2_indeces = [], []
    
    for i in range(len(seq1)):
        
        if seq1[i] != '-' and seq2[i] != '-':
            
            seq1_indeces.append( subst_model.alphabet_dict[seq1[i]] )
            seq2_indeces.append( subst_model.alphabet_dict[seq2[i]] )

    return np.asarray([seq1_indeces, seq2_indeces], dtype=int)     


def p_distance(seq1, seq2, exclude_gaps=True):
    """Calculate the p-distance of two aligned sequences.
    
    Equals the (length-)normalized Hamming distance.
    
    Paramaters
    ----------
    seq1 : str
        Sequence 1.
    seq2 : str
        Sequence 2.
    exclude_gaps : bool, optional
        Ignore columns with a gap in one sequence, gaps in both sequences are
        always ignored. The default is True.
    """
    
    if len(seq1) != len(seq2):
        raise ValueError(f'unequal sequence lengths: {len(seq1)} and {len(seq2)}')
        
    diffs, valid_columns = 0, 0
    
    for i in range(len(seq1)):
        
        if seq1[i] != '-' and seq2[i] != '-':
            
            valid_columns += 1
            if seq1[i] != seq2[i]:
                diffs += 1
        
        elif not exclude_gaps and (seq1[i] != '-' or seq2[i] != '-'):
            
            valid_columns += 1
            diffs += 1
            
    p = diffs / valid_columns if valid_columns > 0 else float('inf')
    
    return p, valid_columns


def JC69_distance(seq1, seq2, exclude_gaps=True, amino_acid=False, 
                  variance=False):
    """Jukes Cantor 1969 distance."""
    
    p, valid_columns = p_distance(seq1, seq2, exclude_gaps=exclude_gaps)
    
    d = _JC69_transform(p, amino_acid=amino_acid)
    
    if not variance:
        return d
    
    else:
        var_d = _JC69_distance_var(p, valid_columns,
                                   amino_acid=amino_acid)
        return d, var_d


def _JC69_transform(p, amino_acid=False):
    """Jukes Cantor 1969 transformation of p-distance."""
    
    a = 3 if not amino_acid else 19     # numerator
    b = 4 if not amino_acid else 20     # denominator
    
    if p >= a / b:
        d = float('nan')
    
    else:
        d = - (a/b) * np.log(1 - (b/a) * p)
        
    return d


def _JC69_distance_var(p, n, amino_acid=False):
    """Jukes Cantor 1969 distance variance."""
    
    a = 3 if not amino_acid else 19     # numerator
    b = 4 if not amino_acid else 20     # denominator
    
    if p >= a / b:
        var = float('nan')
    
    else:
        var = p * (1 - p) / (n * (1 - (b/a) * p) ** 2)
        
    return var


def K80_distance(seq1, seq2, variance=False):
    """Kimura 1980 distance and transition/transversion ratio."""
    
    S, V, valid_columns = _IV_proportions(seq1, seq2)
    
    d, kappa = _K80_transform(S, V)
    
    if not variance:
        return d, kappa
    
    else:
        var_d = _K80_distance_var(S, V, valid_columns)
        return d, kappa, var_d
    

def _IV_proportions(seq1, seq2):
    """Computes the proportions of transitions and transversions."""
    
    purines = {'A', 'a', 'G', 'g', 'R', 'r'}
    pyrimidines = {'C', 'c', 'T', 't', 'U', 'u', 'Y', 'y'}
    
    if len(seq1) != len(seq2):
        raise ValueError(f'unequal sequence lengths: {len(seq1)} and {len(seq2)}')
        
    transitions, transversions, valid_columns = 0, 0, 0
    
    for i in range(len(seq1)):
        
        if seq1[i] != '-' and seq2[i] != '-':
            
            valid_columns += 1
            
            if seq1[i] == seq2[i]:
                continue
            
            elif ((seq1[i] in purines and seq2[i] in purines) or
                  (seq1[i] in pyrimidines and seq2[i] in pyrimidines)):
                transitions += 1
            
            elif ((seq1[i] in purines and seq2[i] in pyrimidines) or
                  (seq1[i] in pyrimidines and seq2[i] in purines)):
                transversions += 1
                
            
    S = transitions / valid_columns if valid_columns > 0 else float('nan')
    V = transversions / valid_columns if valid_columns > 0 else float('nan')
    
    return S, V, valid_columns


def _K80_transform(S, V):
    """Kimura 1980 distance and transition/transversion ratio."""
    
    a1 = 1.0 - 2.0 * S - V
    a2 = 1.0 - 2.0 * V
    
    if a1 <= 0.0 or a2 <= 0.0:
        return float('nan'), float('nan')
    
    else:
        d = - 0.5 * np.log(a1) - 0.25 * np.log(a2)
        kappa = 2.0 * np.log(a1) / np.log(a2) - 1.0 if a2 < 1.0 else float('nan')
        
        return d, kappa


def _K80_distance_var(S, V, n):
    """Kimura 1980 distance variance."""
    
    a1 = 1.0 - 2.0 * S - V
    a2 = 1.0 - 2.0 * V
    
    if a1 <= 0.0 or a2 <= 0.0:
        return float('nan')
    
    else:
        a = 1.0 / a1
        b = 0.5 * ( 1.0 / a1 + 1.0 / a2 )
        
        return (a**2 * S + b**2 * V - (a * S + b * V)**2) / n



# -*- coding: utf-8 -*-

import numpy as np
from scipy.special import zeta


__author__ = 'David Schaller'


class Sampler:
    """Sampling from a variety of distributions."""
    
    def __init__(self, params, minimum=None, maximum=None, discrete=False,
                 shift=0):
        """
        Parameters
        ----------
        params : float or tuple
            The distribution and its parameters, see documentation for
            available options.
        minimum : float, optional
            Minimum value to be sampled. The default is None.
        maximum : float, optional
            Maximum value to be sampled. The default is None.
        discrete : bool, optional
            If True, round sampled values in case of a continuous
            distribution. The default is False.
        shift : float, optional
            Shift distribution along x-axis. The default is 0, i.e., the
            distribution is not shifted.
        """
        
        self._params = params
        self._discrete = discrete
        self._min = minimum
        self._max = maximum
        self._shift = shift
        
        # function to be defined in _check_distribution
        self.draw = None
        
        self._check_distribution()
    
    
    def __call__(self):
        """Draw a value from the specified distribution."""
        
        return self.draw()
        
    
    def _check_distribution(self):
        
        if self._min is not None and not isinstance(self._min, (int, float)):
            raise ValueError('minimum must be a number')
        if self._max is not None  and not isinstance(self._max, (int, float)):
            raise ValueError('maximum must be a number')
        if not isinstance(self._shift, (int, float)):
            raise ValueError('shift value must be a number')
            
        
        if isinstance(self._params, (int, float)):
            
            self._distr = 'constant'
            self.draw = self._draw_constant
            self._exp_val = (round(self._params) if self._discrete else
                             self._params)
            
        elif isinstance(self._params, (tuple, list)) and len(self._params) >= 2:
            
            if self._params[0] == 'constant':
                
                val = self._params[1]
                if not isinstance(val, (int, float)):
                    raise ValueError('constant value must be a number')
                    
                self._distr = 'constant'
                self.draw = self._draw_constant
                self._exp_val = round(val) if self._discrete else val
                
            elif self._params[0] == 'uniform':
                
                a = self._params[1]
                if not len(self._params) >= 3:
                    raise ValueError('uniform distr. requires 2 parameters')
                b = self._params[2]
                
                if (not isinstance(a, (int, float)) or 
                    not isinstance(b, (int, float)) or b < a):
                    raise ValueError('parameters a and b for uniform distr.'\
                                     'must be numbers and a <= b')
                    
                self._distr = 'uniform'
                self.draw = self._draw_uniform
                self._exp_val = (a + b) / 2
                self._a = a
                self._b = b
                
            elif self._params[0] == 'discrete_uniform':
                
                a = self._params[1]
                if not len(self._params) >= 3:
                    raise ValueError('discrete uniform distr. requires 2 '\
                                     'parameters')
                b = self._params[2]
                
                if (not isinstance(a, int) or 
                    not isinstance(b, int) or b < a):
                    raise ValueError('parameters a and b for discrete uniform '\
                                     'distr. must be ints and a <= b')
                    
                self._distr = 'discrete_uniform'
                self.draw = self._draw_discrete_uniform
                self._exp_val = (a + b) / 2
                self._a = a
                self._b = b
                    
            elif self._params[0] == 'gamma':
                
                shape = self._params[1]
                if not len(self._params) >= 3:
                    raise ValueError('gamma distr. requires 2 parameters')
                scale = self._params[2]
                
                if (not isinstance(shape, float) or shape <= 0.0 or 
                    not isinstance(scale, float) or scale <= 0.0):
                    raise ValueError('scale and shape parameters for gamma '\
                                     'distribution must be floats >0.0')
                    
                self._distr = 'gamma'
                self.draw = self._draw_gamma
                self._exp_val = shape * scale
                self._shape = shape
                self._scale = scale
            
            elif self._params[0] == 'gamma_mean':
                
                mean = self._params[1]
                
                if not isinstance(mean , (int, float)) or mean <= 0.0:
                    raise ValueError('mean of gamma distribution must be '\
                                     'a number >0.0')
                
                self._distr = 'gamma'
                self.draw = self._draw_gamma
                self._exp_val = mean
                self._shape = 1.0
                self._scale = mean / self._shape
                
                
            elif self._params[0] == 'exponential':
                
                rate = self._params[1]
                
                if not isinstance(mean , float) or rate < 0.0:
                    raise ValueError('rate must be a float >=0.0')
                
                self._distr = 'exponential'
                self.draw = self._draw_exponential
                self._exp_val = float('inf') if rate == 0.0 else 1/rate
                self._rate = rate
                
                
            elif self._params[0] == 'zipf':
                
                a = self._params[1]
                    
                if not isinstance(a, float) or a <= 1.0:
                    raise ValueError('parameter a for zipf distr. must be a '\
                                     'float >1')
                
                # expected value is infinite for a <= 2
                if a > 2.0:
                    self._exp_val = zeta(a - 1.0) / zeta(a)
                    
                # use some valid value when exp. value is infinite
                elif self._min is not None:
                    self._exp_val = self._min
                else:
                    self._exp_val = 0.0
                    
                self._distr = 'zipf'
                self.draw = self._draw_zipf
                self._a = a
                    
            elif self._params[0] == 'negative_binomial':
                
                r = self._params[1]
                if not len(self._params) >= 3:
                    raise ValueError('negative binomial distr. requires 2 '\
                                     'parameters')
                q = self._params[2]
                    
                if not isinstance(r, int) or r < 1:
                    raise ValueError('parameter r must be an int and >0')
                if not isinstance(q, float) or q <= 0 or q >= 1.0:
                    raise ValueError('parameter q must be a float >0 and <1')
                    
                self._distr = 'negative_binomial'
                self.draw = self._draw_neg_bin
                self._exp_val = r * q / (1 - q)
                self._r = r
                self._q = q
                
            else:
                raise ValueError("length distribution '{}' is not "\
                                 "supported".format(self._params[0]))
            
        else:
            raise ValueError('could not parse distribution')
        
        self._exp_val += self._shift
        
        # expected value must be between min. and max.
        if ((self._min is not None and self._exp_val < self._min) or
            (self._max is not None and self._exp_val > self._max)):
            raise ValueError('expected value must be >= minimum and'\
                             '<= maximum')
            
                
    def _draw_constant(self):
        
        return self._exp_val
    
    
    def _draw_uniform(self):
        
        while True:
            x = np.random.uniform(low=self._a, high=self._b) + self._shift
            
            if self._discrete:
                x = round(x)
                
            if ((not self._min or x >= self.min) and
                (not self._max or x <= self._max)):
                return x
            
    
    def _draw_discrete_uniform(self):
        
        while True:
            x = np.random.randint(self._a, high=self._b) + self._shift
                
            if ((not self._min or x >= self.min) and
                (not self._max or x <= self._max)):
                return x
                
    
    def _draw_gamma(self):
        
        while True:
            x = np.random.gamma(self._shape, scale=self._scale) + self._shift
            
            if self._discrete:
                x = round(x)
                
            if ((not self._min or x >= self.min) and
                (not self._max or x <= self._max)):
                return x
            
            
    def _draw_exponential(self):
        
        while True:
            if self._rate == 0.0:
                x = float('inf')
            else:
                x = np.random.exponential(scale=1/self._rate) + self._shift
            
            if self._discrete:
                x = round(x)
                
            if ((not self._min or x >= self.min) and
                (not self._max or x <= self._max)):
                return x
            
    
    def _draw_zipf(self):
        
        while True:
            x = np.random.zipf(self._a) + self._shift
                
            if ((not self._min or x >= self.min) and
                (not self._max or x <= self._max)):
                return x
            
    
    def _draw_neg_bin(self):
        
        while True:
            x = np.random.negative_binomial(self._r, 1 - self._q) + self._shift
                
            if ((not self._min or x >= self.min) and
                (not self._max or x <= self._max)):
                return x



# -*- coding: utf-8 -*-


__author__ = 'David Schaller'


def parse_paml(filename, model_type='a'):
    
    if model_type == 'a':
        N = 20
    elif model_type == 'n':
        N = 4
    else:
        raise ValueError("model type '{}' not supported".format(model_type))
    
    with open(filename, 'r') as f:
        
        exchangeability_matrix = [[0.0 for j in range(N)] for i in range(N)]
        
        line = f.readline().strip()
        while line == '': line = f.readline().strip()
        
        for i in range(1,N):
            
            line = [float(item) for item in line.split()]
            
            for j in range(len(line)):
                exchangeability_matrix[i][j] = line[j]
                exchangeability_matrix[j][i] = line[j]
                
            line = f.readline().strip()
                
        line = f.readline().strip()
        while line == '': line = f.readline().strip()
            
        stat_freqs = [float(item) for item in line.split()]
        
        if len(stat_freqs) != N:
            raise RuntimeError('wrong no. of equilibrium frequencies, '\
                               'check paml file')
        
        return exchangeability_matrix, stat_freqs


# -*- coding: utf-8 -*-

"""Indel models for sequence evolution.

References
----------
.. [1] Z. Yang.
   Computational molecular evolution.
   Oxford series in ecology and evolution. Oxford University Press, 2006.
   ISBN 978-0-19-856699-1 978-0-19-856702-8.
.. [2] R. A. Cartwright.
   DNA assembly with gaps (Dawg): Simulating sequence evolution.
   In: Bioinformatics, 21(Suppl 3):iii31-iii38, November 2005.
   doi:10.1093/bioinformatics/bti1200.
"""

from asymmetree.tools.Sampling import Sampler


__author__ = 'David Schaller'


class IndelModel:
    """Indel model"""
    
    def __init__(self, insertion_rate, deletion_rate,
                 length_distr=('zipf', 1.821),        # (Chang and Benner 2004)
                 min_length=1,
                 max_length=None,
                 **kwargs):
        """
        Parameters
        ----------
        insertition_rate : float
            Insertion rate w.r.t. units of evolutionary distance.
        deletion_rate : float
            Deletion rate w.r.t. units of evolutionary distance.
        length_distr : tuple, optional
            Distribution of the length of the indels, see documentation for
            available options. The default is a Zipf distribution as observed
            for empirical data in [1].
        min_length : int, optional
            The minimal length of an indel, default is 1.
        max_length : int, optional
            The maximal length of an indel. The default is None in which case
            the distribution is not truncated (on the upper side).
        
        References
        ----------
        .. [1] M. S. Chang and S. A. Benner.
           Empirical Analysis of Protein Insertions and Deletions Determining
           Parameters for the Correct Placement of Gaps in Protein Sequence
           Alignments. 
           In: Journal of Molecular Biology, 341(2):617-631, August 2004.
           doi:10.1016/j.jmb.2004.05.045.
        """
        
        if insertion_rate < 0.0 or deletion_rate < 0.0:
            raise ValueError("insertion and deletion rates must be non-negative")
        else:
            self._ins_rate = insertion_rate     # insertion rate per site
            self._del_rate = deletion_rate      # deletion rate per site
        
        if isinstance(length_distr, Sampler):
            self.sampler = length_distr
        else:
            self.sampler = Sampler(length_distr,
                                   maximum=max_length,
                                   discrete=True)
        
    
    def get_rates(self, seq_length):
        """Return the current insertion and deletion rate.
        
        Computes the current insertion and deletion rate according to the model
        and the length of the sequence.
        
        Parameters
        ----------
        seq_length : int
            The length of the sequence.
        
        Returns
        -------
        tuple of two floats
            The total rate for an insertation and deletion event, see [1].
            
        References
        ----------
        .. [1] R. A. Cartwright.
           DNA assembly with gaps (Dawg): Simulating sequence evolution.
           In: Bioinformatics, 21(Suppl 3):iii31-iii38, November 2005.
           doi:10.1093/bioinformatics/bti1200.
        """
        
        # expected value may be infinite for zipf distribution, in which case
        # the minimal length is used, see 'Sampler'
        return ( (seq_length + 1) * self._ins_rate,
                 (seq_length + self.sampler._exp_val - 1) * self._del_rate )
        
    
    def draw_length(self):
        """Draw the length for an indel.
        
        Returns
        -------
        int
            The drawn length for the indel.
        """
        
        return self.sampler()
    


# -*- coding: utf-8 -*-

"""HGT, xenology and Fitch graphs.

This module contains several functions for the analysis of horizontal gene
transfer events in the simulated scenarios. In particular, the directed and
undirected Fitch graph can be extracted, as well as the pairs of genes that
diverged later than the respective species in which they reside, i.e. the
later-divergence-time (LDT) graph, see [1]. The latter situation is indicative 
or the presence of HGT events in the scenario.

References
----------
.. [1] D. Schaller, M. Lafond, P.F. Stadler, N. Wieseke, M. Hellmuth.
   Indirect identification of horizontal gene transfer.
   In: Journal of Mathematical Biology, 2021, 83(1):10.
   doi: 10.1007/s00285-021-01631-0.
"""

import itertools

import networkx as nx

from tralda.datastructures import Tree, TreeNode, LCA
from tralda.tools.GraphTools import independent_sets
from tralda.cograph.Cograph import (to_cotree, paths_of_length_2)
from tralda.supertree import Build

from asymmetree.tools.PhyloTreeTools import (add_planted_root,
                                             phylo_tree_attributes,
                                             assign_missing_labels)
from asymmetree.treeevolve.SpeciesTree import distance_from_timing


__author__ = 'David Schaller'


# --------------------------------------------------------------------------
#                     Transfer edges and Fitch graphs
# --------------------------------------------------------------------------

def true_transfer_edges(T):
    """Returns a set containing v if (u, v) is labeled as a transfer edge.
    
    Parameters
    ----------
    T : Tree
        A tree whose nodes have the 'transferred' attribute.
    
    Returns
    -------
    set
        The subset of TreeNode instances in the tree for which transferred is 1.
    """
    
    return {v for _, v in T.edges() if v.transferred}


def rs_transfer_edges(T, S, lca_S=None):
    """Transfer edges in T according to the relaxed scenario definition.
    
    An edge (u,v) in T is an (rs-)transfer edge if u and v are mapped to
    incomparable nodes/edges in the species tree S.
    
    Parameters
    ----------
    T : Tree
        The gene tree.
    S : Tree
        The species tree corresponding to the gene tree.
    lca_S : tralda.datastructures.Tree.LCA, optional
        Instance of LCA corresponding to the species tree, default is None
        in which case a new instance is created and used.
    
    Returns
    -------
    set
        The subset of TreeNode instances in the tree for which the 'reconc'
        attributes of its incident edges refer to incomparable branches in the
        species tree.
    """
    
    if not isinstance(lca_S, LCA):
        lca_S = LCA(S)
        
    transfer_edges = set()
    
    for u, v in T.edges():
        if not lca_S.are_comparable(u.reconc, v.reconc):
            transfer_edges.add(v)
    
    return transfer_edges


def fitch(tree, transfer_edges, supply_undirected=False, lca_T=None):
    """Returns the (directed) Fitch graph.
    
    Parameters
    ----------
    tree : Tree
        The gene tree.
    transfer_edges : iterable
        The subset of TreeNode instances in the tree that are transfer edges.
    supply_undirected : bool, optional
        If True, additionally return the undirected Fitch graph, default is
        False.
    lca_T : tralda.datastructures.Tree.LCA, optional
        Instance of LCA corresponding to the tree, default is False in which
        case a new instance is created and used.
        
    Returns
    -------
    networkx.DiGraph or tuple
        The directed Fitch graph or a tuple containing the directed Fitch graph
        (networkx.DiGraph) and the undirected Fitch graph (networkx.Graph).
    """
    
    if not isinstance(lca_T, LCA):
        lca_T = LCA(tree)
        
    if not isinstance(transfer_edges, (set, dict)):
        transfer_edges = set(transfer_edges)
    
    leaves = [l for l in tree.leaves()]
    fitch = nx.DiGraph()
    
    # store for each leaf the first transfer edge on the way to the root
    first_transfer = {}
    
    for x in leaves:
        fitch.add_node(x.label, color=x.reconc)
        
        current = x
        while current:
            if current in transfer_edges:
                first_transfer[x] = current
                break
            current = current.parent
    
    for x, y in itertools.permutations(leaves, 2):
        
        if (y in first_transfer and
            lca_T.ancestor_not_equal(lca_T(x, y), first_transfer[y])):
            fitch.add_edge(x.label, y.label)
    
    if not supply_undirected:
        return fitch
    else:
        return fitch, fitch.to_undirected()
    

def undirected_fitch(tree, transfer_edges, lca_T=None):
    """Returns the undirected Fitch graph.
    
    Parameters
    ----------
    tree : Tree
        The gene tree.
    transfer_edges : iterable
        The subset of TreeNode instances in the tree that are transfer edges.
    lca_T : tralda.datastructures.Tree.LCA, optional
        Instance of LCA corresponding to the tree, default is False in which
        case a new instance is created and used.
        
    Returns
    -------
    networkx.Graph
        The undirected Fitch graph (networkx.Graph).
    """
    
    return fitch(tree, transfer_edges, supply_undirected=True, lca_T=lca_T)[1]


def _rs_fitch_aux_graph(G, color_set, ind_sets, ignore_set):
    
    aux_graph = nx.Graph()
    aux_graph.add_nodes_from(color_set)
    
    for i, ind_set in enumerate(ind_sets):
        
        # skip the independent set that shall not be included
        if ignore_set == i:
            continue
        
        for x, y in itertools.combinations(ind_set, 2):
            X = G.nodes[x]['color']
            Y = G.nodes[y]['color']
            
            if X != Y:
                aux_graph.add_edge(X, Y)
    
    return aux_graph


def is_rs_fitch(G, color_set=None):
    """Checks whether a given graph is an rs-Fitch graph.
    
    I.e. whether the graph is the Fitch graph of some relaxed scenario (rs),
    see [1].
    
    Parameters
    ----------
    G : networkx.Graph
        A graph whose nodes have the 'color' attribute.
    color_set : set, optional
        The set of colors appearing in the graph. The default is None, in which
        case such a set is created internally.
    
    Returns
    -------
    bool
        True if the graph is an rs-Fitch graph as defined in [1].
        
    References
    ----------
    .. [1] D. Schaller, M. Lafond, P.F. Stadler, N. Wieseke, M. Hellmuth.
       Indirect identification of horizontal gene transfer.
       In: Journal of Mathematical Biology, 2021, 83(1):10.
       doi: 10.1007/s00285-021-01631-0.
    """
    
    if color_set is None:
        color_set = set()
        for v in G.nodes():
            color_set.add(G.nodes[v]['color'])
    
    ind_sets = independent_sets(G)
    
    # not a complete multipartite graph
    if ind_sets is False:
        return False
    
    # trivial number of independent sets
    k = len(ind_sets)
    if k <= 1:
        return True
    
    # more than one independent set
    for ignore_set in range(-1, k):
        
        # ignore_set = -1 means that no set is ignored
        aux_graph = _rs_fitch_aux_graph(G, color_set, ind_sets, ignore_set)
        if not nx.is_connected(aux_graph):
            return True
    
    # if nothing was True
    return False


# --------------------------------------------------------------------------
#                   Later-divergence-time (LDT) graphs
# --------------------------------------------------------------------------

def below_equal_above(T, S, lca_T=None, lca_S=None):
    """Detemine the pairs of genes that diverged later, at the same time, or 
    earlier as their species.
    
    Returns three graphs with the leaves of the gene tree T as vertex set, and 
    edges ab if and only if a and b diverged later, at the same time, or later,
    resp., than the corresponding species A and B in the species tree S.
    
    Parameters
    ----------
    T : Tree
        The gene tree.
    S : Tree
        The corresponding species tree.
    lca_T : tralda.datastructures.Tree.LCA, optional
        Instance of LCA corresponding to the gene tree, default is False in
        which case a new instance is created and used.
    lca_S : tralda.datastructures.Tree.LCA, optional
        Instance of LCA corresponding to the species tree, default is False in
        which case a new instance is created and used.
    
    Returns
    -------
    tuple of three networkx.Graph instances
        The below, equal, and above relation as described above.
    """
    
    L_T = [l for l in T.leaves()]
    L_S = {l.label: l for l in S.leaves()}
    
    below = nx.Graph()
    equal = nx.Graph()
    above = nx.Graph()
    for u in L_T:
        below.add_node(u.label, color=u.reconc)
        equal.add_node(u.label, color=u.reconc)
        above.add_node(u.label, color=u.reconc)
    
    if not isinstance(lca_T, LCA):
        lca_T = LCA(T)
    if not isinstance(lca_S, LCA):
        lca_S = LCA(S)
    
    for a, b in itertools.combinations(L_T, 2):
        
        t_ab = lca_T.get(a, b).tstamp
        t_AB = lca_S.get(L_S[a.reconc], L_S[b.reconc]).tstamp
        
        if t_ab < t_AB:
            below.add_edge(a.label, b.label)
        elif t_ab == t_AB:
            equal.add_edge(a.label, b.label)
        else:
            above.add_edge(a.label, b.label)
    
    
    return below, above, equal


def ldt_graph(T, S, lca_T=None, lca_S=None):
    """Later-divergence-time graph, see [1].
    
    Returns a graph with the leaves of the gene tree T as vertex set, and 
    edges ab if and only if a and b diverged later than the corresponding
    species A and B in the species tree S.
    
    Parameters
    ----------
    T : Tree
        The gene tree.
    S : Tree
        The corresponding species tree.
    lca_T : tralda.datastructures.Tree.LCA, optional
        Instance of LCA corresponding to the gene tree, default is False in
        which case a new instance is created and used.
    lca_S : tralda.datastructures.Tree.LCA, optional
        Instance of LCA corresponding to the species tree, default is False in
        which case a new instance is created and used.
    
    References
    ----------
    .. [1] D. Schaller, M. Lafond, P.F. Stadler, N. Wieseke, M. Hellmuth.
       Indirect identification of horizontal gene transfer.
       In: Journal of Mathematical Biology, 2021, 83(1):10.
       doi: 10.1007/s00285-021-01631-0.
    """
    
    return below_equal_above(T, S, lca_T=lca_T, lca_S=lca_S)[0]


def _simulate_timing(tree):
    """Simulates a timing for the tree.
    
    It is t(root) = 1 and t(x) = 0 for x in L(S)."""
    
    max_depth = {}
    
    for v in tree.postorder():
        if not v.children:
            max_depth[v] = 0
        else:
            max_depth[v] = max(max_depth[c] for c in v.children) + 1
    
    for v in tree.preorder():
        if not v.children:
            v.tstamp = 0.0
        elif not v.parent:
            v.tstamp = 1.0
        else:
            v.tstamp = v.parent.tstamp * (1 - 1 / (max_depth[v]+1))


class RsScenarioConstructor:
    """Construct an rs-scenario for a given LDT graph.
    
    Implementation of the algorithm presented in [1].
    
    References
    ----------
    .. [1] D. Schaller, M. Lafond, P.F. Stadler, N. Wieseke, M. Hellmuth.
       Indirect identification of horizontal gene transfer.
       In: Journal of Mathematical Biology, 2021, 83(1):10.
       doi: 10.1007/s00285-021-01631-0.
    """
    
    def __init__(self, colored_cograph, color_set=None):
        """
        Parameters
        ----------
        colored_cograph : networkx.Graph
            A cograph whose nodes have the 'color' attribute.
        color_set : set, optional
            The set of colors appearing in the graph. The default is None,
            in which case such a set is created internally.
        """
        
        self.G = colored_cograph
        
        self.L = [x for x in self.G.nodes()]
        
        if color_set is None:
            color_set = set()
            for v in self.L:
                color_set.add(self.G.nodes[v]['color'])
        elif not isinstance(color_set, (set, dict)):
            color_set = set(color_set)
        self.color_set = color_set
        
        
    def run(self):
        """Construct an rs-scenario consisting of a species and gene tree.
        
        Returns
        -------
        tuple of two Tree instances or bool
            The species and gene tree, or False if the supplied graph was 
            not an LDT graph.
        """
        
        self.S = self._species_tree()
        if not self.S:
            return False
        
        self.epsilon = float('inf')
        for u, v in self.S.edges():
            self.epsilon = min(self.epsilon,
                               abs(u.tstamp - v.tstamp))
        self.epsilon /= 3.0
        
        self.S_leaves = self.S.leaf_dict()
        
        # top-level call on full leaf set and rho_S
        self.T = Tree(self._build_gene_tree(self.L, self.S.root.children[0]))
        
        # --- post-processing of T ---
        
        # planted root
        add_planted_root(self.T)
        self.T.root.tstamp = self.S.root.tstamp
        self.T.root.reconc = self.S.root.label
        
        # suppress all vertices with a single child except the planted root
        to_suppress = []
        for u in self.T.preorder():
            if u.parent and len(u.children) == 1:
                to_suppress.append((u.parent, u))
        self.T.contract(to_suppress)
        
        distance_from_timing(self.T)
        
        return self.S, self.T
        
    
    def _species_tree(self):
        
        cotree = to_cotree(self.G)
        
        if not cotree:
            return False
        
        triples = set()
        
        for a, b, c in paths_of_length_2(cotree):
            
            A = self.G.nodes[a.label]['color']
            B = self.G.nodes[b.label]['color']
            C = self.G.nodes[c.label]['color']
            
            if A != B and A != C and B != C:
                
                # sort to avoid redundancy in triple set
                if A > C:
                    A, C = C, A
                
                triples.add( (A, C, B) )
        
        build = Build(triples, self.color_set, mincut=False)
        S = build.build_tree()
        
        if not S:
            return False
        else:
            phylo_tree_attributes(S, inplace=True)
            add_planted_root(S)
            assign_missing_labels(S)
            _simulate_timing(S)
            distance_from_timing(S)
            return S
    
    
    def _build_gene_tree(self, L, u_S):
        
        u_T = TreeNode(label='', event='D',
                       reconc=(u_S.parent.label, u_S.label),
                       tstamp=u_S.tstamp + self.epsilon)
        
        # u_S is a leaf
        if not u_S.children:
            for x in L:
                leaf = TreeNode(label=x,
                                reconc=self.G.nodes[x]['color'],
                                tstamp=0.0)
                u_T.add_child(leaf)
        
        # u_S is an inner vertex
        else:
            
            # maps color to the respective child of u_S
            color_to_v_S = {}
            for v_S in u_S.children:
                for x in self.S_leaves[v_S]:
                    color_to_v_S[x.label] = v_S
                    
            # connected components C of G[L']
            for C in nx.connected_components(self.G.subgraph(L)):
                
                # choose v_S_star s.t. L(S(v_S_star)) \cap sigma(C) is non-empty
                v_S_star = color_to_v_S[self.G.nodes[next(iter(C))]['color']]
                v_T = TreeNode(label='', event='H',
                               reconc=(u_S.label, v_S_star.label),
                               tstamp=u_S.tstamp - self.epsilon)
                u_T.add_child(v_T)
                
                # aux. graph for equivalence classes
                aux_graph = nx.Graph()
                aux_graph.add_nodes_from(C)
                
                for x, y in itertools.combinations(C, 2):
                    if (color_to_v_S[self.G.nodes[x]['color']] is
                        color_to_v_S[self.G.nodes[y]['color']]):
                        aux_graph.add_edge(x, y)
                
                # for each equivalence class K that is a subset of C
                for K in nx.connected_components(aux_graph):
                    
                    # choose v_S s.t. \sigma(K) \subseteq L(S(v_S))
                    v_S = color_to_v_S[self.G.nodes[next(iter(K))]['color']]
                    
                    w_K = self._build_gene_tree(K, v_S)
                    v_T.add_child(w_K)
                    
                    if v_S is not v_S_star:
                        w_K.transferred = 1
        
        return u_T
    
    
# --------------------------------------------------------------------------
#                  Compatibility of trees and partitions
# --------------------------------------------------------------------------

def is_compatible(T, partition, lca=None):
    """Checks whether a tree and a partition are compatible.
    
    A tree with leaf set L and a partition of L are compatible if there is a
    subset of tree edges such that the forest obtained by their removal induces
    the given partition (in terms of membership to the same connected
    component).
    In this case, the vertices can be uniquely colored by set A of the
    partition if they lie on a path connecting two elements from A.
    
    Parameters
    ----------
    T : Tree
        A tree with unique leaf label.
    partition : list or tuple
        A partition of the labels of the trees' leaves.
    lca : LCA, optional
        Precomputed LCA datastructure for the tree. The default is None in
        which case it is computed internally.
    
    Returns
    -------
    bool
        Returns False if the tree and the partition are incompatible, and
        otherwise a tuple of a dict and a list.
    dict
        A dictionary containing vertices of the tree that lie on a path
        connecting two elements from the same set of the partition with the
        index of this set as value.
    list
        A list containing the last common ancestor in the tree for each set
        in the partition.
    """
    
    if not isinstance(T, Tree):
        raise TypeError("T must be of type 'Tree'")
    
    if not isinstance(partition, (list, tuple)):
        raise TypeError("partition must be of type 'list' or 'tuple'")
    
    if not lca:
        lca = LCA(T)
    
    # map labels to leaves of the tree
    label_to_leaf = {v.label: v for v in T.leaves()}
    
    # color each vertex if lies on the path connecting two vertices from the
    # same set
    vertex_coloring = {}
    
    # last common ancestors of the sets in the partition
    lcas = []
    
    for i, A in enumerate(partition):
        
        A_iterator = iter(A)
        x_1 = label_to_leaf[next(A_iterator)]
        current_lca = x_1
        visited = {x_1}
        vertex_coloring[x_1] = i
        
        for x_label in A_iterator:
            x = label_to_leaf[x_label]
            new_lca = lca(x, current_lca)
            
            starts = [x]
            if new_lca is not current_lca:
                starts.append(current_lca.parent)
                current_lca = new_lca
            
            for current in starts:
            
                while current:
                    if current in visited:
                        break
                    
                    if current in vertex_coloring and \
                        vertex_coloring[current] != i:
                        return False
                    else:
                        vertex_coloring[current] = i
                        visited.add(current)
                    
                    if current is new_lca:
                        break
                    
                    current = current.parent
        
        lcas.append(current_lca)
    
    return vertex_coloring, lcas


def is_refinement_compatible(T, partition, lca=None):
    """Checks whether a tree and a partition are refinement-compatible.
    
    A tree T with leaf set L and a partition of L are refinement-compatible if
    there is a refinement T' of T for which a subset of edges of T' exists
    such that the forest obtained by their removal induces the given partition
    (in terms of membership to the same connected component).
    In this case, the edges can be uniquely colored by set A of the partition
    if they lie on a path connecting two elements from A.
    
    Parameters
    ----------
    T : Tree
        A tree with unique leaf label.
    partition : list or tuple
        A partition of the labels of the trees' leaves.
    lca : LCA, optional
        Precomputed LCA datastructure for the tree. The default is None in
        which case it is computed internally.
    
    Returns
    -------
    bool
        Returns False if the tree and the partition are not refinement-
        compatible, and otherwise a tuple of a dict and a list.
    dict
        A dictionary containing nodes v of the tree such that the edge
        (parent(v), v) lies on a path connecting two elements from the same
        set of the partition with the index of this set as value.
    list
        A list containing the last common ancestor in the tree for each set
        in the partition.
    """
    
    if not isinstance(T, Tree):
        raise TypeError("T must be of type 'Tree'")
    
    if not isinstance(partition, (list, tuple)):
        raise TypeError("partition must be of type 'list' or 'tuple'")
    
    if not lca:
        lca = LCA(T)
    
    # map labels to leaves of the tree
    label_to_leaf = {v.label: v for v in T.leaves()}
    
    # color each vertex v if the edge (v.parent, v) lies on the path
    # connecting two vertices from the same set
    edge_coloring = {}
    
    # last common ancestors of the sets in the partition
    lcas = []
    
    for i, A in enumerate(partition):
        
        A_iterator = iter(A)
        current_lca = label_to_leaf[next(A_iterator)]
        visited = set()
        
        for x_label in A_iterator:
            x = label_to_leaf[x_label]
            new_lca = lca(x, current_lca)
            
            starts = [x]
            if new_lca is not current_lca:
                starts.append(current_lca)
                current_lca = new_lca
            
            for current in starts:
            
                while current:
                    
                    if current in edge_coloring and \
                        edge_coloring[current] != i:
                        return False
                    else:
                        edge_coloring[current] = i
                        visited.add(current)
                        
                    # parent is always defined since at least the first edge
                    # must be added and if the root is reached, it must equal
                    # new_lca
                    current = current.parent
                    if current in visited or current is new_lca:
                        break
        
        lcas.append(current_lca)
    
    return edge_coloring, lcas


def fitch_orientation(T, partition, lca=None):
    """(Un)ambiguously present or absent edges in the directed Fitch graph.
    
    The ordered pair (x, y) of two leaves x and y is in the Fitch relation
    if the path from lca(x, y) to y contains a transfer edge. An undirected
    Fitch relation ((x, y) if (x, y) or (y, x) in the directed relation) can
    be represented by a partition (the maximal independent sets).
    In combination with a tree, this partition determines to some extent the
    directed Fitch relation.
    
    Parameters
    ----------
    T : Tree
        A tree with unique leaf label.
    partition : list or tuple
        A partition of the labels of the trees' leaves.
    lca : LCA, optional
        Precomputed LCA datastructure for the tree. The default is None in
        which case it is computed internally.
    
    Returns
    -------
    bool
        Returns False if the tree and the partition are incompatible.
    2-dimensional array (list of lists)
        Stores at indices i and j corresponding to sets A and B, resp., of the
        partition whether (A, B) is 'essential', 'forbidden', or 'ambiguous'.
    """
    
    if not isinstance(T, Tree):
        raise TypeError("T must be of type 'Tree'")
    
    if not isinstance(partition, (list, tuple)):
        raise TypeError("partition must be of type 'list' or 'tuple'")
    
    if not lca:
        lca = LCA(T)
    
    compatibility = is_compatible(T, partition, lca=lca)
    if compatibility:
        vertex_coloring, lcas = compatibility
    else:
        return False
    
    # compute for each vertex its lowest colored (strict) ancestor
    lowest_colored_ancestor = {}
    for v in T.preorder():
        if not v.parent:
            lowest_colored_ancestor[v] = None
        elif v.parent in vertex_coloring:
            lowest_colored_ancestor[v] = v.parent
        else:
            lowest_colored_ancestor[v] = lowest_colored_ancestor[v.parent]
    
    n = len(partition)
    matrix = [[None for j in range(n)] for i in range(n)]
    
    for i in range(n):
        for j in range(n):
            
            if i == j:
                matrix[i][j] = 'forbidden'
                continue
            
            u = lca(lcas[i], lcas[j])       # lca(lca(A), lca(B))
            w = lowest_colored_ancestor[lcas[j]]
            
            # lca(B) < u and there is a colored vertex w with lca(B) < w <= u
            if (lca.ancestor_not_equal(u, lcas[j]) and
                w and 
                lca(w, u) is u):
                matrix[i][j] = 'essential'
            # lca(A) < lca(B)
            elif lca.ancestor_not_equal(lcas[j], lcas[i]):
                matrix[i][j] = 'forbidden'
            else:
                matrix[i][j] = 'ambiguous'
    
    return matrix


def _find_child(u, v, lca):
    """For v < u, find the child w of u such that v <= w."""
    
    for w in u.children:
        if lca(v, w) is w:
            return w
    
    raise RuntimeError('could not find corresponding child of u '\
                       '(u not an ancestor of v?)')


def fitch_orientation_for_refinements(T, partition, lca=None):
    """(Un)ambiguously present or absent edges in the directed Fitch graph of
    all compatible refinements.
    
    The ordered pair (x, y) of two leaves x and y is in the Fitch relation
    if the path from lca(x, y) to y contains a transfer edge. An undirected
    Fitch relation ((x, y) if (x, y) or (y, x) in the directed relation) can
    be represented by a partition (the maximal independent sets).
    In combination with a tree, this partition determines to some extent the
    directed Fitch relation w.r.t. the possible refinements of the tree.
    
    Parameters
    ----------
    T : Tree
        A tree with unique leaf label.
    partition : list or tuple
        A partition of the labels of the trees' leaves.
    lca : LCA, optional
        Precomputed LCA datastructure for the tree. The default is None in
        which case it is computed internally.
    
    Returns
    -------
    bool
        Returns False if the tree and the partition are not refinement-
        compatible.
    2-dimensional array (list of lists)
        Stores at indices i and j corresponding to sets A and B, resp., of the
        partition whether (A, B) is 'essential', 'forbidden', or 'ambiguous'
        for all refinements of the tree that ae compatible with the partition.
    """
    
    if not isinstance(T, Tree):
        raise TypeError("T must be of type 'Tree'")
    
    if not isinstance(partition, (list, tuple)):
        raise TypeError("partition must be of type 'list' or 'tuple'")
    
    if not lca:
        lca = LCA(T)
    
    r_compatibility = is_refinement_compatible(T, partition, lca=lca)
    if r_compatibility:
        edge_coloring, lcas = r_compatibility
    else:
        return False
    
    # compute for each vertex its lowest colored (strict) ancestor
    lowest_colored_edge = {}
    for v in T.preorder():
        if not v.parent:
            lowest_colored_edge[v] = None
        elif v in edge_coloring:
            lowest_colored_edge[v] = v
        else:
            lowest_colored_edge[v] = lowest_colored_edge[v.parent]
    
    n = len(partition)
    matrix = [[None for j in range(n)] for i in range(n)]
    
    for i in range(n):
        for j in range(n):
            
            if i == j:
                matrix[i][j] = 'forbidden'
                continue
            
            u = lca(lcas[i], lcas[j])       # lca(lca(A), lca(B))
            v_A = lcas[i]                   # lca(A)
            v_B = lcas[j]                   # lca(B)
            lce_B = lowest_colored_edge[v_B]
            
            if (lca.ancestor_not_equal(u, v_B) and
                lce_B and lca.ancestor_not_equal(u, lce_B)):
                matrix[i][j] = 'essential'
            elif (u.parent and
                  lca.ancestor_not_equal(u, v_A) and
                  u in edge_coloring and
                  edge_coloring[u] == \
                      edge_coloring.get(_find_child(u, v_A, lca))):
                matrix[i][j] = 'essential'
            elif (lca.ancestor_not_equal(v_B, v_A) and
                  edge_coloring.get(_find_child(u, v_A, lca)) == j):
                matrix[i][j] = 'forbidden'
            else:
                matrix[i][j] = 'ambiguous'
    
    return matrix
    



# -*- coding: utf-8 -*-

"""AsymmeTree - Simulation of complex gene families

AsymmeTree is an open-source Python library for the simulation and analysis of
phylogenetic scenarios. It includes a simulator for species and gene trees with
heterogeneous evolution rates, nucleotide and amino acid sequences with or
without indels, as well as whole genomes/proteomes.

Moreover, it includes a matplotlib-based visualization of the simulated trees
as well as tools for the extraction of information from the simulated scenarios
such as orthology, best matches, and xenology.

The library is primarily designed to explore and validate mathematical concepts,
and to test inference methods for various steps on the way to more
realistically-available data, i.e., dated gene trees, additive distances of
gene sets, noisy distances and finally sequences.
"""



# -*- coding: utf-8 -*-

"""
Data structure for evolving sequences.
"""

from enum import Enum, auto

from tralda.datastructures.DoublyLinkedList import (DLList,
                                                    DLListNode,
                                                    DLListIterator)


__author__ = 'David Schaller'


class State(Enum):
    """State of the sites in evolving sequences."""
    
    ROOT       = auto()
    INSERTION  = auto()
    INHERITED  = auto()


class EvoSeqNode(DLListNode):
    """Site in an evolving sequence."""
    
    __slots__ = ('status', 'site_id', 'parent_el', 'rate_class', 'rate_factor')
    
    def __init__(self, value,
                 status, site_id,
                 prev_node=None, next_node=None,
                 parent_el=None,
                 rate_class=0, rate_factor=1.0):
        
        super().__init__(value, prev_node=prev_node, next_node=next_node)
        
        self.status = status
        self.site_id = site_id
        self.parent_el = parent_el
        self.rate_class = rate_class
        self.rate_factor = rate_factor


class EvoSeq(DLList):
    """Evolving sequence."""
    
    def __init__(self):
        
        super().__init__()
        
        
    def __iter__(self):
        
        return EvoSeqIterator(self)


    def __next__(self):
        
        pass
    
    
    def append(self, value, status, site_id, parent_el=None,
               rate_class=0, rate_factor=1.0):
        
        new_end = EvoSeqNode(value, status, site_id,
                             prev_node=self._last,
                             parent_el=parent_el,
                             rate_class=rate_class,
                             rate_factor=rate_factor)
        if self._last:
            self._last._next = new_end
        self._last = new_end
        if not self._first:
            self._first = new_end
        self._count += 1
        return new_end
    
    
    def append_left(self, value, status, site_id, parent_el=None,
                    rate_class=0, rate_factor=1.0):
        
        new_start = EvoSeqNode(value, status, site_id,
                               next_node=self._first,
                               parent_el=parent_el,
                               rate_class=rate_class,
                               rate_factor=rate_factor)
        if self._first:
            self._first._prev = new_start
        self._first = new_start
        if not self._last:
            self._last = new_start
        self._count += 1
        return new_start
    
    
    def insert_right_of(self, element, value, status, site_id, parent_el=None,
                        rate_class=0, rate_factor=1.0):
        
        if element is self._last:
            new_element = self.append(value, status, site_id,
                                      parent_el=parent_el,
                                      rate_class=rate_class,
                                      rate_factor=rate_factor)
            
        else:
            new_element = EvoSeqNode(value, status, site_id,
                                     prev_node=element,
                                     next_node=element._next,
                                     parent_el=parent_el,
                                     rate_class=rate_class,
                                     rate_factor=rate_factor)
            new_element._next._prev = new_element
            element._next = new_element
            self._count += 1
        
        return new_element    
        
    
    def clone(self):
        
        child_seq = EvoSeq()
        
        parent_site = self._first
        while parent_site:
            child_seq.append(parent_site._value,
                             State.INHERITED, parent_site.site_id,
                             parent_el=parent_site,
                             rate_class=parent_site.rate_class,
                             rate_factor=parent_site.rate_factor)
            parent_site = parent_site._next
            
        return child_seq
    
    
    def element_pairs(self):
        """Generator for subsequent element pairs."""
        
        if self._count >= 2:
            
            first = self._first
            second = self._first._next
            
            while second:
                yield (first, second)
                first = second
                second = second._next
    
    
    def count_status(self, status):
        
        counter = 0
        
        site = self._first
        while site:
            if site.status == status:
                counter += 1
            site = site._next
                
        return counter
        
          
            
class EvoSeqIterator(DLListIterator):
    """Iterator class for evolving sequence."""
    
    def __init__(self, evoseq):
        
        super().__init__(evoseq)
        
    
    def __next__(self):
        """Next site element in the evolving sequence.
        
        Overrides the super class methods and returns the element instead
        of the value only."""
        
        if self._current:
            x = self._current
            self._current = self._current._next
            return x
        else:
            raise StopIteration